
"""
Job Execution Script
"""

import os
import getpass
import platform
import crocodile.toolbox as tb
from crocodile.cluster.loader_runner import JobParams, FileManager, WorkloadParams, JOB_STATUS
from crocodile.cluster.remote_machine import RemoteMachineConfig
from importlib.machinery import SourceFileLoader
from rich.console import Console
from rich.panel import Panel
from rich import inspect
from rich.text import Text
import pandas as pd

console = Console()

# EXTRA-PLACEHOLDER-PRE

_ = SourceFileLoader, WorkloadParams

params = JobParams.from_empty()

print("\n" * 2)
manager: FileManager = FileManager.from_pickle(params.file_manager_path)
manager.secure_resources()
pid: int = os.getpid()
manager.execution_log_dir.expanduser().joinpath("pid.txt").create(parents_only=True).write_text(str(pid))
job_status: JOB_STATUS = "running"
manager.execution_log_dir.expanduser().joinpath("status.txt").write_text(job_status)


# keep those values after lock is released
time_at_execution_start_utc = pd.Timestamp.utcnow()
time_at_execution_start_local = pd.Timestamp.now()
manager.execution_log_dir.expanduser().create().joinpath("start_time.txt").write_text(str(time_at_execution_start_local))
func_kwargs = tb.Read.vanilla_pickle(path=manager.kwargs_path.expanduser())

# EXTRA-PLACEHOLDER-POST


# ######################### EXECUTION ####################################

print("\n" * 2)
console.rule(title="PYTHON EXECUTION SCRIPT", style="bold red", characters="-")
print("\n" * 2)
console.print(f"Executing {tb.P(rf'{params.repo_path_rh}').expanduser().collapseuser().as_posix()}/{params.file_path_r} : {params.func_name}", style="bold blue")
if isinstance(func_kwargs, dict): tb.S(func_kwargs).print(title="kwargs", as_config=True)
else: inspect(func_kwargs, value=False, title=f"kwargs from `{manager.kwargs_path.collapseuser().as_posix()}`", docs=False, sort=False)
print("\n" * 2)

res = ""
func = ""

# execution_line

print("\n" * 2)
console.rule(title="FINISHED PYTHON EXECUTION SCRIPT", characters="-", style="bold red")
print("\n" * 2)

# ######################### END OF EXECUTION #############################


if type(res) is tb.P or (type(res) is str and tb.P(res).expanduser().exists()):
    res_folder = tb.P(res).expanduser()
else:
    res_folder = tb.P.tmp(folder=rf"tmp_dirs/{manager.job_id}").create()
    console.print(Panel(f"WARNING: The executed function did not return a path to a results directory. Execution metadata will be saved separately in {res_folder.collapseuser().as_posix()}."))
    print("\n\n")
    # try:
        # tb.Save.pickle(obj=res, path=res_folder.joinpath("result.pkl"))
    # except TypeError as e:
        # print(e)
        # print(f"Could not pickle res object to path `{res_folder.joinpath('result.pkl').collapseuser().as_posix()}`.")

time_at_execution_end_utc = pd.Timestamp.utcnow()
time_at_execution_end_local = pd.Timestamp.now()
delta = time_at_execution_end_utc - time_at_execution_start_utc
exec_times = tb.S({"start_utc 🌍⏲️": time_at_execution_start_utc, "end_utc 🌍⏰": time_at_execution_end_utc,
                   "start_local ⏲️": time_at_execution_start_local, "end_local ⏰": time_at_execution_end_local, "delta ⏳": delta,
                   "submission_time": manager.submission_time, "wait_time": time_at_execution_start_local - manager.submission_time})

# save the following in results folder and execution log folder.:
manager.execution_log_dir.expanduser().joinpath("end_time.txt").write_text(str(time_at_execution_end_local))
manager.execution_log_dir.expanduser().joinpath("results_folder_path.txt").write_text(res_folder.collapseuser().as_posix())
manager.execution_log_dir.expanduser().joinpath("error_message.txt").write_text(params.error_message)
exec_times.save(path=manager.execution_log_dir.expanduser().joinpath("execution_times.Struct.pkl"))
if params.error_message == "":
    job_status = "completed"
    manager.execution_log_dir.expanduser().joinpath("status.txt").write_text(job_status)
else:
    job_status = "failed"
    manager.execution_log_dir.expanduser().joinpath("status.txt").write_text(job_status)
print(f"job {manager.job_id} is completed.")


tb.Experimental.generate_readme(path=manager.job_root.expanduser().joinpath("execution_log.md"), obj=func, desc=f'''

Job executed via tb.cluster.Machine
remote: {params.ssh_repr}
job_id: {manager.job_id}

py_script_path @ `{manager.py_script_path.collapseuser()}`
shell_script_path @ `{manager.shell_script_path.collapseuser()}`
kwargs_path @ `{manager.kwargs_path.collapseuser()}`

### Execution Time:
{exec_times.print(as_config=True, return_str=True)}

### Job description
{params.description}

''')


# manager.root_dir.expanduser().copy(folder=res_folder, overwrite=True)

# print to execution console:
exec_times.print(title="Execution Times", as_config=True)
print("\n" * 1)
ssh_repr_remote = params.ssh_repr_remote or f"{getpass.getuser()}@{platform.node()}"  # os.getlogin() can throw an error in non-login shells.
console.print(Panel(Text(f'''
ftprx {ssh_repr_remote} {res_folder.collapseuser()} -r
''', style="bold blue on white"), title="Pull results with this line:", border_style="bold red"))


if params.session_name != "":
    if platform.system() == "Linux":
        tb.Terminal().run(f"""zellij --session {params.session_name} action new-tab --name results  """)
        # --layout ~/code/machineconfig/src/machineconfig/settings/zellij/layouts/d.kdl --cwd {res_folder.as_posix()}
        tb.Terminal().run(f"""zellij --session {params.session_name} action write-chars "cd {res_folder.as_posix()};lf" """)
    elif platform.system() == "Windows":
        tb.Terminal().run(f"""wt --window {params.session_name} new-tab --title results -startingDirectory {res_folder.as_posix()} lf """)


# NOTIFICATION-CODE-PLACEHOLDER


manager.unlock_resources()
rm_conf: RemoteMachineConfig = tb.Read.vanilla_pickle(path=manager.remote_machine_config_path.expanduser())


if rm_conf.kill_on_completion:
    # assert rm_conf.launch_method == "cloud_manager"
    if platform.system() == "Linux":
        from crocodile.cluster.session_managers import Zellij  # type: ignore  # pylint: disable=C0412
        current_session = Zellij.get_current_zellij_session()
        # Zellij.close_tab(sess_name=params.session_name, tab_name=params.tab_name)
        print(f"Killing session `{params.session_name}` on `{params.ssh_repr}`")
        tb.Terminal().run(f"zellij --session {current_session} go-to-tab-name '{params.tab_name}'; sleep 2; zellij --session {current_session} action close-tab").print()  # i.e. current tab
    elif platform.system() == "Windows":
        print(f"Killing session `{params.session_name}` on `{params.ssh_repr}`")
        from machineconfig.utils.procs import ProcessManager
        pm = ProcessManager()
        pm.kill(commands=[params.session_name])
    else: raise NotImplementedError(f"kill_on_completion is not implemented for platform `{platform.system()}`")
else:
    print(f"Keeping the tab `{params.tab_name}` on `{params.ssh_repr}`")


console.rule(title="END OF PYTHON EXECUTION SCRIPT", style="bold red", characters="-")




"""Odds
"""

from crocodile.file_management import P, install_n_import
from crocodile.core import List
# from crocodile.meta import Scheduler, Log
# import time
from typing import Optional, Any, Callable, TypeVar, Generic


def share(path: P):
    response = install_n_import("requests").post(url=r'https://file.io', data=dict(name=path.name, expires="2022-08-01", title="a descriptivefile title", maxDownloads=1, autoDelete=True, private=False, description="its a file, init?", ), files={'file': path.read_bytes()})
    return response.json()['link'] if ['link'] in response.json() else response.json()
def edit_video(path: P, t_start: float = 0.0, t_end: Optional[float] = None, speed: float = 1.0, suffix: Optional[str] = None, rotate: float = 0.0, volume: float = 1.0, fps: float = 25.0):
    ed = install_n_import("moviepy").editor
    clip = ed.VideoFileClip(path); print(f"{clip.size=}, {clip.duration=}, {clip.fps=}")
    clip.subclip(t_start=t_start, t_end=t_end).rotate(rotate).volumex(volume).fx(ed.vfx.speedx, speed).write_videofile(path.append("_modified").with_suffix(path.suffix if suffix is None else suffix), fps=fps)


def capture_from_webcam(show: bool = True, wait: bool = True, save: bool = False):
    cv2 = install_n_import("cv2", "opencv-python")
    cam = cv2.VideoCapture(0)
    cam.set(cv2.CAP_PROP_FRAME_WIDTH, 3000)
    cam.set(cv2.CAP_PROP_FRAME_HEIGHT, 2000)
    _ret, frame = cam.read()
    while True and wait:
        _ret, frame = cam.read()
        if show: cv2.imshow('frame', frame)
        if cv2.waitKey(1) & 0xFF == ord('q'): break
    cam.release()
    cv2.destroyAllWindows()
    if save:
        path = P.tmpfile(suffix=".jpg")
        cv2.imwrite(str(path), frame)
        return path
    return frame


def qr(txt: str): install_n_import("qrcode").make(txt).save((file := P.tmpfile(suffix=".png")).__str__()); return file()
def count_number_of_lines_of_code_in_repo(path: P = P.cwd(), extension: str = ".py", r: bool = True, **kwargs: Any): return P(path).search(f"*{extension}", r=r, **kwargs).read_text(encoding="utf-8").splitlines().apply(len).np.sum()
def profile_memory(command: str):
    psutil = install_n_import("psutil")
    before = psutil.virtual_memory()
    exec(command)  # type: ignore # pylint: disable=W0122
    after = psutil.virtual_memory()
    print(f"Memory used = {(after.used - before.used) / 1e6}")


T = TypeVar('T')


class Cycle(Generic[T]):
    def __init__(self, iterable: list[T]):
        self.list = iterable
        self.index, self.prev_index = 0, -1
    def next(self) -> T: self.prev_index = self.index; self.index += 1; self.index = 0 if self.index >= len(self.list) else self.index; return self.list[self.index]
    def previous(self) -> T: self.prev_index = self.index; self.index -= 1; self.index = len(self.list) - 1 if self.index < 0 else self.index; return self.list[self.index]
    def set_value(self, value: T): self.prev_index = self.index; self.index = self.list.index(value)
    def get_value(self) -> T: return self.list[self.index]
    def get_index(self) -> int: return self.index
    def set_index(self, index: int): self.prev_index = self.index; self.index = index
    def expand(self, val: T): self.list.append(val)
    def __add__(self, other: 'Cycle[T]'): pass  # see behviour of matplotlib cyclers.
    def __repr__(self): return f"Cycler @ {self.index}: {self.list[self.index]}"
class DictCycle(Cycle[tuple[str, T]]):
    def __init__(self, strct: dict[str, T]):
        super(DictCycle, self).__init__(iterable=list(strct.items()))
        self.keys = strct.keys()
    def set_key(self, key: Any): self.index = list(self.keys).index(key)


def tree(self: P, level: int = -1, limit_to_directories: bool = False, length_limit: int = 1000, stats: bool = False, desc: Optional[Callable[[Any], str]] = None):
    # Based on: https://stackoverflow.com/questions/9727673/list-directory-tree-structure-in-python"""
    space, branch, tee, last, dir_path, files, directories = '    ', '│   ', '├── ', '└── ', self, 0, 0
    def get_stats(apath: P): return (f" {(sts := apath.stats())['size']} MB. {sts['content_mod_time']}. " + desc(apath) if desc is not None else "") if stats or desc else ""
    def inner(apath: P, prefix: str = '', level_: int = -1):
        nonlocal files, directories
        if not level_: return  # 0, stop iterating
        pointers = [tee] * (len(content := apath.search("*", files=not limit_to_directories)) - 1) + [last]
        for pointer, path in zip(pointers, content):
            if path.is_dir():
                yield prefix + pointer + path.name + get_stats(path)
                directories, extension = directories + 1, branch if pointer == tee else space
                yield from inner(path, prefix=prefix + extension, level_=level_ - 1)
            elif not limit_to_directories: yield prefix + pointer + path.name + get_stats(path); files += 1
    print(dir_path.name)
    iterator = inner(dir_path, level_=level)
    _ = [print(line) for line in __import__("itertools").islice(iterator, length_limit)]
    _ = print(f'... length_limit, {length_limit}, reached, counted:') if next(iterator, None) else None
    print(f'\n{directories} directories' + (f', {files} files' if files else ''))


def get_compressable_directories(path: P, max_size_mb: float = 15_000.0):
    tmp_results = path.search("*", r=False)
    dirs2compress = List()
    dirs_violating = List()
    files_violating = List()
    for item in tmp_results:
        if item.size() > max_size_mb:  # should be parsed
            if item.is_file():
                print(f"File `{item}` has size larger than maximum allowed, consider manual handling")
                files_violating.append(item)
            else:
                tmp_dirs2compress, tmp_dirs_violating, tmp_files_violating = get_compressable_directories(item, max_size_mb=max_size_mb)
                if len(tmp_dirs2compress) == 0:
                    print(f"Directory `{item}` has size larger than maximum allowed, but when parsed for subdirectories to be compressed, nothing could be found. Handle manually")
                    dirs_violating.append(item)
                dirs2compress += tmp_dirs2compress
                dirs_violating += tmp_dirs_violating
                files_violating += tmp_files_violating
        else:
            if item.is_dir() and len(item.search("*", folders=False, r=True)) > 10:  # too many files ==> compress
                dirs2compress.append(item)
    return dirs2compress, dirs_violating, files_violating


if __name__ == '__main__':
    pass




"""
dl
"""

import crocodile.toolbox as tb
from crocodile.file_management import P, Path
from crocodile.matplotlib_management import ImShow
# from matplotlib.pyplot import hist
import numpy as np
import numpy.typing as npt
import pandas as pd
from abc import ABC
from typing import TypeVar, Type, Any, Optional, Union, Callable, Literal
import enum
from tqdm import tqdm
import copy
from dataclasses import dataclass, field


@dataclass
class Specs:
    ip_names: list[str]  # e.g.: ["x1", "x2"]
    op_names: list[str]  # e.g.: ["y1", "y2"]
    other_names: list[str] = field(default_factory=list)  # e.g.: indices or names
    ip_shapes: list[tuple[int, ...]] = field(default_factory=list)
    op_shapes: list[tuple[int, ...]] = field(default_factory=list)
    other_shapes: list[tuple[int, ...]] = field(default_factory=list)
    def get_all_strings(self): return self.ip_names + self.op_names + self.other_names
    def get_split_strings(self, strings: list[str], which_split: Literal["train", "test"] = "train") -> list[str]:
        keys_ip = [item + f"_{which_split}" for item in strings]
        return keys_ip


@dataclass
class EvaluationData:
    x: list[Any]
    y_pred: list[Any]
    y_pred_pp: Any
    y_true: list[Any]
    y_true_pp: list[Any]
    names: list[str]
    loss_df: Optional['pd.DataFrame']
    def __repr__(self) -> str:
        print("")  # this is useful to move to new line in IPython console and skip the header `In [5]` which throws off table aliegnment of header and content.
        _ = tb.S(self.__dict__).print()
        return "EvaluationData Object"


@dataclass
class DeductionResult:
    input: 'npt.NDArray[np.float64]'
    preprocessed: 'npt.NDArray[np.float64]'
    postprocessed: 'npt.NDArray[np.float64]'
    prediction: 'npt.NDArray[np.float64]'


class Device(enum.Enum):
    gpu0 = 'gpu0'
    gpu1 = 'gpu1'
    cpu = 'cpu'
    two_gpus = '2gpus'
    auto = 'auto'


SubclassedHParams = TypeVar("SubclassedHParams", bound='HParams')
SubclassedDataReader = TypeVar("SubclassedDataReader", bound='DataReader')
SubclassedBaseModel = TypeVar("SubclassedBaseModel", bound='BaseModel')


PRECISON = Literal['float64', 'float32', 'float16']


@dataclass
class HParams:
    # ===================== Data ==============================
    seed: int
    shuffle: bool
    precision: PRECISON
    # ===================== Model =============================
    # depth = 3
    # ===================== Training ==========================
    test_split: float  # test split
    learning_rate: float
    batch_size: int
    epochs: int
    # ================== General ==============================
    name: str  # field(default_factory=lambda: "model-" + tb.randstr(noun=True))
    root: tb.P  # = tb.P.tmp(folder="tmp_models")
    # _configured: bool = False
    # device_na: None = None
    pkg_name: Literal['tensorflow', 'torch'] = 'tensorflow'
    device_name: Device = Device.gpu0
    subpath: str = 'metadata/hyperparameters'  # location within model directory where this will be saved.

    def save(self):
        self.save_dir.joinpath(self.subpath, 'hparams.txt').create(parents_only=True).write_text(str(self))
        try: data: dict[str, Any] = self.__getstate__()
        except AttributeError:
            data = self.__dict__
        tb.Save.vanilla_pickle(path=self.save_dir.joinpath(self.subpath, "hparams.HParams.dat.pkl"), obj=data)
        tb.Save.vanilla_pickle(path=self.save_dir.joinpath(self.subpath, "hparams.HParams.pkl"), obj=self)
    def __getstate__(self) -> dict[str, Any]: return self.__dict__
    def __setstate__(self, state: dict[str, Any]): return self.__dict__.update(state)
    @classmethod
    def from_saved_data(cls, path: Union[str, Path, P], *args: Any, **kwargs: Any):
        data: dict[str, Any] = tb.Read.vanilla_pickle(path=tb.P(path) / cls.subpath / "hparams.HParams.dat.pkl", *args, **kwargs)
        return cls(**data)
    # def __repr__(self, **kwargs: Any): return "HParams Object with specs:\n" + tb.Struct(self.__dict__).print(as_config=True, return_str=True)
    @property
    def pkg(self):
        if self.pkg_name not in ("tensorflow", "torch"): raise ValueError(f"pkg_name must be either `tensorflow` or `torch`")
        return __import__("tensorflow") if self.pkg_name == "tensorflow" else __import__("torch")
    @property
    def save_dir(self) -> tb.P: return (tb.P(self.root) / self.name).create()


class DataReader:
    subpath = tb.P("metadata/data_reader")
    """This class holds the dataset for training and testing.
    """
    def __init__(self, hp: SubclassedHParams,  # type: ignore
                 specs: Optional[Specs] = None,
                 split: Optional[dict[str, Any]] = None) -> None:
        # split could be Union[None, 'npt.NDArray[np.float64]', 'pd.DataFrame', 'pd.Series', 'list[Any]', Tf.RaggedTensor etc.
        super().__init__()
        self.hp = hp
        self.split = split if split is not None else {}
        self.plotter = None
        self.specs: Specs = Specs(ip_shapes=[], op_shapes=[], other_shapes=[], ip_names=[], op_names=[], other_names=[]) if specs is None else specs
        # self.df_handler = df_handler
    def save(self, path: Optional[str] = None, **kwargs: Any) -> None:
        _ = kwargs
        base = (tb.P(path) if path is not None else self.hp.save_dir).joinpath(self.subpath).create()
        try: data: dict[str, Any] = self.__getstate__()
        except AttributeError: data = self.__dict__
        tb.Save.vanilla_pickle(path=base / "data_reader.DataReader.dat.pkl", obj=data)
        tb.Save.vanilla_pickle(path=base / "data_reader.DataReader.pkl", obj=self)
    @classmethod
    def from_saved_data(cls, path: Union[str, tb.P], hp: SubclassedHParams,  # type: ignore
                        **kwargs: Any):
        path = tb.P(path) / cls.subpath / "data_reader.DataReader.dat.pkl"
        data: dict[str, Any] = tb.Read.vanilla_pickle(path)
        obj = cls(hp=hp, **kwargs)
        obj.__setstate__(data)
        return obj
    def __getstate__(self) -> dict[str, Any]:
        items: list[str] = ["specs"]
        res = {}
        for item in items:
            if hasattr(self, item): res[item] = getattr(self, item)
        return res
    def __setstate__(self, state: dict[str, Any]) -> None: return self.__dict__.update(state)
    def __repr__(self):
        print(f"DataReader Object with these keys: \n")
        tb.S(self.specs.__dict__).print(as_config=True, title="Data Specs")  # config print
        if bool(self.split):
            print("Split-Data Table:")
            tb.S(self.split).print(as_config=False, title="Split Data")  # table print
        return f"--" * 50

    def split_the_data(self, data_dict: dict[str, Any], populate_shapes: bool, split_kwargs: Optional[dict[str, Any]] = None) -> None:
        # populating data specs ip / op shapes based on arguments sent to this method.
        strings = self.specs.get_all_strings()
        keys = list(data_dict.keys())
        if len(strings) != len(keys) or set(keys) != set(strings):
            tb.S(self.specs.__dict__).print(as_config=True, title="Specs Declared")
            # tb.S(data_dict).print(as_config=True, title="Specs Declared")
            print(f"data_dict keys: {keys}")
            raise ValueError(f"Arguments mismatch! The specs that you declared have keys that do not match the keys of the data dictionary passed to split method.")

        if populate_shapes:
            delcared_ip_shapes = self.specs.ip_shapes
            delcared_op_shapes = self.specs.op_shapes
            delcared_other_shapes = self.specs.other_shapes
            self.specs.ip_shapes = []
            self.specs.op_shapes = []
            self.specs.other_shapes = []
            for data_name, data_value in data_dict.items():
                a_shape = data_value.iloc[0].shape if type(data_value) in {pd.DataFrame, pd.Series} else np.array(data_value[0]).shape
                if data_name in self.specs.ip_names: self.specs.ip_shapes.append(a_shape)
                elif data_name in self.specs.op_names: self.specs.op_shapes.append(a_shape)
                elif data_name in self.specs.other_names: self.specs.other_shapes.append(a_shape)
                else: raise ValueError(f"data_name `{data_name}` is not in the specs. I don't know what to do with it.\n{self.specs=}")
            if len(delcared_ip_shapes) != 0 and len(delcared_op_shapes) != 0:
                if delcared_ip_shapes != self.specs.ip_shapes or delcared_op_shapes != self.specs.op_shapes or delcared_other_shapes != self.specs.other_shapes:
                    print(f"Declared ip shapes: {delcared_ip_shapes}")
                    print(f"Declared op shapes: {delcared_op_shapes}")
                    print(f"Declared other shapes: {delcared_other_shapes}")
                    print(f"Populated ip shapes: {self.specs.ip_shapes}")
                    print(f"Populated op shapes: {self.specs.op_shapes}")
                    print(f"Populated other shapes: {self.specs.other_shapes}")
                    raise ValueError(f"Shapes mismatch! The shapes that you declared do not match the shapes of the data dictionary passed to split method.")

        from sklearn.model_selection import train_test_split  # type: ignore
        args = [data_dict[item] for item in strings]
        result = train_test_split(*args, test_size=self.hp.test_split, shuffle=self.hp.shuffle, random_state=self.hp.seed, **split_kwargs if split_kwargs is not None else {})
        self.split = {}  # dict(train_loader=None, test_loader=None)
        self.split.update({astring + '_train': result[ii * 2] for ii, astring in enumerate(strings)})
        self.split.update({astring + '_test': result[ii * 2 + 1] for ii, astring in enumerate(strings)})
        print(f"================== Training Data Split ===========================")
        tb.Struct(self.split).print()
        print(f"==================================================================")

    def sample_dataset(self, aslice: Optional['slice'] = None, indices: Optional[list[int]] = None,
                       use_slice: bool = False, split: Literal["train", "test"] = "test", size: Optional[int] = None) -> tuple[list[Any], list[Any], list[Any]]:
        assert self.split is not None, f"No dataset is loaded to DataReader, .split attribute is empty. Consider using `.load_training_data()` method."
        keys_ip = self.specs.get_split_strings(self.specs.ip_names, which_split=split)
        keys_op = self.specs.get_split_strings(self.specs.op_names, which_split=split)
        keys_others = self.specs.get_split_strings(self.specs.other_names, which_split=split)

        tmp = self.split[keys_ip[0]]
        assert tmp is not None, f"Split key {keys_ip[0]} is None. Make sure that the data is loaded."
        ds_size = len(tmp)
        select_size = size or self.hp.batch_size
        start_idx = np.random.choice(ds_size - select_size)

        selection: Union[list[int], slice]
        if indices is not None: selection = indices
        elif aslice is not None: selection = list(range(aslice.start, aslice.stop, aslice.step))
        elif use_slice: selection = slice(start_idx, start_idx + select_size, 1)  # ragged tensors don't support indexing, this can be handy in that case.
        else:
            tmp2: list[int] = np.random.choice(ds_size, size=select_size, replace=False).astype(int).tolist()
            selection = tmp2
        x: list[Any] = []
        y: list[Any] = []
        others: list[Any] = []
        for idx, key in zip([0] * len(keys_ip) + [1] * len(keys_op) + [2] * len(keys_others), keys_ip + keys_op + keys_others):
            tmp3: Any = self.split[key]
            if isinstance(tmp3, (pd.DataFrame, pd.Series)):
                item = tmp.iloc[np.array(selection)]
            elif tmp3 is not None:
                item = tmp3[selection]
            elif tmp3 is None: raise ValueError(f"Split key {key} is None. Make sure that the data is loaded.")
            else: raise ValueError(f"Split key `{key}` is of unknown data type `{type(tmp3)}`.")
            if idx == 0: x.append(item)
            elif idx == 1: y.append(item)
            else: others.append(item)
        # x = x[0] if len(self.specs.ip_names) == 1 else x
        # y = y[0] if len(self.specs.op_names) == 1 else y
        # others = others[0] if len(self.specs.other_names) == 1 else others
        # if len(others) == 0:
        #     if type(selection) is slice: others = np.arange(*selection.indices(10000000000000)).tolist()
        #     else: others = selection
        return x, y, others

    def get_random_inputs_outputs(self, ip_shapes: Optional[list[tuple[int, ...]]] = None, op_shapes: Optional[list[tuple[int, ...]]] = None):
        if ip_shapes is None: ip_shapes = self.specs.ip_shapes
        if op_shapes is None: op_shapes = self.specs.op_shapes
        dtype = self.hp.precision if hasattr(self.hp, "precision") else "float32"
        x = [np.random.randn(self.hp.batch_size, * ip_shape).astype(dtype) for ip_shape in ip_shapes]
        y = [np.random.randn(self.hp.batch_size, * op_shape).astype(dtype) for op_shape in op_shapes]
        # x = x[0] if len(self.specs.ip_names) == 1 else x
        # y = y[0] if len(self.specs.op_names) == 1 else y
        return x, y

    def preprocess(self, *args: Any, **kwargs: Any): _ = args, kwargs, self; return args[0]  # acts like identity.
    def postprocess(self, *args: Any, **kwargs: Any): _ = args, kwargs, self; return args[0]  # acts like identity

    # def standardize(self):
    #     assert self.split is not None, "Load up the data first before you standardize it."
    #     self.scaler = StandardScaler()
    #     self.split['x_train'] = self.scaler.fit_transform(self.split['x_train'])
    #     self.split['x_test']= self.scaler.transform(self.split['x_test'])

    def image_viz(self, pred: 'npt.NDArray[np.float64]', gt: Optional[Any] = None, names: Optional[list[str]] = None, **kwargs: Any):
        """
        Assumes numpy inputs
        """
        if gt is None: self.plotter = ImShow(pred, labels=None, sup_titles=names, origin='lower', **kwargs)
        else: self.plotter = ImShow(img_tensor=pred, sup_titles=names, labels=['Reconstruction', 'Ground Truth'], origin='lower', **kwargs)

    def viz(self, eval_data: EvaluationData, **kwargs: Any):
        """Implement here how you would visualize a batch of input and ouput pair. Assume Numpy arguments rather than tensors."""
        _ = self, eval_data, kwargs
        return None


@dataclass
class Compiler:
    loss: Any
    optimizer: Any
    metrics: list[Any]


class BaseModel(ABC):
    """My basic model. It implements the following methods:

    * :func:`BaseModel.preprocess` This should convert to tensors as appropriate for the model.
    * :func:`BaseModel.postprocess` This method should convert back to numpy arrays.
    * :func:`BaseModel.infer` This method expects processed input and only forwards through the model
    * :func:`BaseModel.predict` expects a processed input, uese infer and does postprocessing.
    * :func:`BaseModel.predict_from_s` reads, preprocess, then uses predict method.
    * :func:`BseModel.evaluate` Expects processed input and internally calls infer and postprocess methods.

    Functionally or Sequentually built models are much more powerful than Subclassed models. They are faster, have more features, can be plotted, serialized, correspond to computational graphs etc.
    """
    # @abstractmethod
    def __init__(self, hp: SubclassedHParams, data: SubclassedDataReader,  # type: ignore
                 history: Optional[list[dict[str, Any]]] = None):
        # : Optional[list]
        self.hp = hp  # should be populated upon instantiation.
        self.data = data  # should be populated upon instantiation.
        self.model: Any = self.get_model()  # should be populated upon instantiation.
        self.compiler: Compiler
        self.history = history if history is not None else []  # should be populated in fit method, or loaded up.
        # self.plotter = NullAuto
        # self.fig = None
        # self.kwargs = None
        # self.tmp = None
        __module = self.__class__.__module__
        if __module.startswith('__main__'):
            print("WARNING: Model class is defined in main. Saving the code from the current working directory. Consider importing the model class from a module.")

    def get_model(self):
        raise NotImplementedError
        # pass
    def compile(self, loss: Optional[Any] = None, optimizer: Optional[Any] = None, metrics: Optional[list[Any]] = None, compile_model: bool = True):
        """ Updates compiler attributes. This acts like a setter.
        .. note:: * this method is as good as setting attributes of `compiler` directly in case of PyTorch.
                  * In case of TF, this is not the case as TF requires actual futher different
                    compilation before changes take effect.
        Remember:
        * Must be run prior to fit method.
        * Can be run only after defining model attribute.
        """
        pkg = self.hp.pkg
        if self.hp.pkg_name == 'tensorflow':
            if loss is None: loss = pkg.keras.losses.MeanSquaredError()
            if optimizer is None: optimizer = pkg.keras.optimizers.Adam(self.hp.learning_rate)
            if metrics is None: metrics = []  # [pkg.keras.metrics.MeanSquaredError()]
        elif self.hp.pkg_name == 'torch':
            if loss is None: loss = pkg.nn.MSELoss()
            if optimizer is None: optimizer = pkg.optim.Adam(self.model.parameters(), lr=self.hp.learning_rate)
            if metrics is None: metrics = []  # [tmp.MeanSquareError()]
        else: raise ValueError(f"pkg_name must be either `tensorflow` or `torch`")
        # Create a new compiler object
        self.compiler = Compiler(loss=loss, optimizer=optimizer, metrics=list(metrics))
        # in both cases: pass the specs to the compiler if we have TF framework
        if self.hp.pkg.__name__ == "tensorflow" and compile_model: self.model.compile(**self.compiler.__dict__)

    def fit(self, viz: bool = True, val_sample_weights: Optional['npt.NDArray[np.float64]'] = None, **kwargs: Any):
        assert self.data.split is not None, "Split your data before you start fitting."
        x_train = [self.data.split[item] for item in self.data.specs.get_split_strings(self.data.specs.ip_names, which_split="train")]
        y_train = [self.data.split[item] for item in self.data.specs.get_split_strings(self.data.specs.op_names, which_split="train")]
        x_test = [self.data.split[item] for item in self.data.specs.get_split_strings(self.data.specs.ip_names, which_split="test")]
        y_test = [self.data.split[item] for item in self.data.specs.get_split_strings(self.data.specs.op_names, which_split="test")]

        x_test = x_test[0] if len(x_test) == 1 else x_test
        y_test = y_test[0] if len(y_test) == 1 else y_test
        default_settings: dict[str, Any] = dict(x=x_train[0] if len(x_train) == 1 else x_train,
                                                y=y_train[0] if len(y_train) == 1 else y_train,
                                                validation_data=(x_test, y_test) if val_sample_weights is None else (x_test, y_test, val_sample_weights),
                                                batch_size=self.hp.batch_size, epochs=self.hp.epochs, verbose=1, shuffle=self.hp.shuffle, callbacks=[])
        default_settings.update(kwargs)
        hist = self.model.fit(**default_settings)
        self.history.append(copy.deepcopy(hist.history))  # it is paramount to copy, cause source can change.
        if viz:
            artist = self.plot_loss()
            artist.fig.savefig(str(self.hp.save_dir.joinpath(f"metadata/training/loss_curve.png").append(index=True).create(parents_only=True)))
        return self

    def switch_to_sgd(self, epochs: int = 10):
        assert self.compiler is not None, "Compiler is not initialized. Please initialize the compiler first."
        print(f'Switching the optimizer to SGD. Loss is fixed to {self.compiler.loss}'.center(100, '*'))
        if self.hp.pkg.__name__ == 'tensorflow': new_optimizer = self.hp.pkg.keras.optimizers.SGD(lr=self.hp.learning_rate * 0.5)
        else: new_optimizer = self.hp.pkg.optim.SGD(self.model.parameters(), lr=self.hp.learning_rate * 0.5)
        self.compiler.optimizer = new_optimizer
        return self.fit(epochs=epochs)

    def switch_to_l1(self, epochs: int = 10):
        assert self.compiler is not None, "Compiler is not initialized. Please initialize the compiler first."
        if self.hp.pkg.__name__ == 'tensorflow':
            self.model.reset_metrics()
        print(f'Switching the loss to l1. Optimizer is fixed to {self.compiler.optimizer}'.center(100, '*'))
        if self.hp.pkg.__name__ == 'tensorflow':
            new_loss = self.hp.pkg.keras.losses.MeanAbsoluteError()
        else:
            raise NotImplementedError
            # import crocodile.deeplearning_torch as tmp
            # new_loss = tmp.MeanAbsoluteError()
        self.compiler.loss = new_loss
        return self.fit(epochs=epochs)

    def preprocess(self, *args: Any, **kwargs: Any):
        """Converts an object to a numerical form consumable by the NN."""
        return self.data.preprocess(*args, **kwargs)

    def postprocess(self, *args: Any, **kwargs: Any):
        return self.data.postprocess(*args, **kwargs)
    def __call__(self, *args: Any, **kwargs: Any):
        return self.model(*args, **kwargs)
    def viz(self, eval_data: EvaluationData, **kwargs: Any):
        return self.data.viz(eval_data, **kwargs)
    def save_model(self, directory: Union[str, Path, P]): self.model.save(directory)  # In TF: send only path dir. Save path is saved_model.pb
    def save_weights(self, directory: Union[str, Path, P]):
        self.model.save_weights(P(directory).joinpath(self.model.name))  # TF: last part of path is file path.
    @staticmethod
    def load_model(directory: Union[str, Path, P]): __import__("tensorflow").keras.models.load_model(str(directory))  # path to directory. file saved_model.pb is read auto.
    def load_weights(self, directory: Union[str, Path, P]):
        # assert self.model is not None, "Model is not initialized. Please initialize the model first."
        self.model.load_weights(P(directory).search('*.data*').list[0].__str__().split('.data')[0]).expect_partial()  # requires path to file path.
    def summary(self):
        from contextlib import redirect_stdout
        path = self.hp.save_dir.joinpath("metadata/model/model_summary.txt").create(parents_only=True)
        with open(str(path), 'w', encoding='utf-8') as f:
            with redirect_stdout(f): self.model.summary()
        return self.model.summary()
    def config(self): _ = [print(layer.get_config(), "\n==============================") for layer in self.model.layers]; return None
    def plot_loss(self, *args: Any, **kwargs: Any):
        res = tb.Struct.concat_values(*self.history)
        assert self.compiler is not None, "Compiler is not initialized. Please initialize the compiler first."
        if hasattr(self.compiler.loss, "name"): y_label = self.compiler.loss.name
        else: y_label = self.compiler.loss.__name__
        return res.plot_plt(*args, title="Loss Curve", xlabel="epochs", ylabel=y_label, **kwargs)

    def infer(self, x: Any) -> 'npt.NDArray[np.float64]':
        """ This method assumes numpy input, datatype-wise and is also preprocessed.
        NN is put in eval mode.
        :param x:
        :return: prediction as numpy
        """
        # return self.model(x, training=False)  # Keras automatically handles special layers, can accept dataframes, and always returns numpy.
        # https://stackoverflow.com/questions/64199384/tf-keras-model-predict-results-in-memory-leak
        # https://github.com/tensorflow/tensorflow/issues/44711
        return self.model.predict(x)

    def predict(self, x: Any, **kwargs: Any):
        """This method assumes preprocessed input. Returns postprocessed output. It is useful at evaluation time with preprocessed test set."""
        return self.postprocess(self.infer(x), **kwargs)

    def deduce(self, obj: Any, viz: bool = True, **kwargs: Any) -> DeductionResult:
        """Assumes that contents of the object are in the form of a batch."""
        preprocessed = self.preprocess(obj, **kwargs)
        prediction = self.infer(preprocessed)
        postprocessed = self.postprocess(prediction, **kwargs)
        result = DeductionResult(input=obj, preprocessed=preprocessed, prediction=prediction, postprocessed=postprocessed)
        if viz: self.viz(postprocessed, **kwargs)
        return result

    def evaluate(self, x_test: Optional[list['npt.NDArray[np.float64]']] = None, y_test: Optional[list['npt.NDArray[np.float64]']] = None, names_test: Optional[list[str]] = None,
                 aslice: Optional[slice] = None, indices: Optional[list[int]] = None, use_slice: bool = False, size: Optional[int] = None,
                 split: Literal["train", "test"] = "test", viz: bool = True, viz_kwargs: Optional[dict[str, Any]] = None):
        if x_test is None and y_test is None and names_test is None:
            x_test, y_test, names_test = self.data.sample_dataset(aslice=aslice, indices=indices, use_slice=use_slice, split=split, size=size)
        elif names_test is None and x_test is not None:
            names_test = [str(item) for item in np.arange(len(x_test))]
        else: raise ValueError(f"Either provide x_test and y_test or none of them. Got x_test={x_test} and y_test={y_test}")
        # ==========================================================================
        y_pred_raw = self.infer(x_test)
        if not isinstance(y_pred_raw, list): y_pred = [y_pred_raw]
        else: y_pred = y_pred_raw
        assert isinstance(y_test, list)
        loss_df = self.get_metrics_evaluations(y_pred, y_test)
        y_pred_pp = self.postprocess(y_pred, per_instance_kwargs=dict(name=names_test), legend="Prediction")
        y_true_pp = self.postprocess(y_test, per_instance_kwargs=dict(name=names_test), legend="Ground Truth")
        # if loss_df is not None:
            # if len(self.data.specs.other_names) == 1: loss_df[self.data.specs.other_names[0]] = names_test
            # else:
            #     for val, name in zip(names_test, self.data.specs.other_names): loss_df[name] = val
        # loss_name = results.loss_df.columns.to_list()[0]  # first loss path
        # loss_label = results.loss_df[loss_name].apply(lambda x: f"{loss_name} = {x}").to_list()
        # names: list[str] = [f"{aname}. Case: {anindex}" for aname, anindex in zip(loss_label, names_test)]
        results = EvaluationData(x=x_test, y_pred=y_pred, y_pred_pp=y_pred_pp, y_true=y_test, y_true_pp=y_true_pp, names=[str(item) for item in names_test], loss_df=loss_df)
        if viz:
            self.viz(results, **(viz_kwargs or {}))
        return results

    def get_metrics_evaluations(self, prediction: list['npt.NDArray[np.float64]'], groun_truth: list['npt.NDArray[np.float64]']) -> 'pd.DataFrame':
        # if self.compiler is None: return None
        metrics = [self.compiler.loss] + self.compiler.metrics
        loss_dict: dict[str, list[Any]] = dict()
        for a_metric in metrics:
            if hasattr(a_metric, "name"): name = a_metric.name
            elif hasattr(a_metric, "__name__"): name = a_metric.__name__
            elif hasattr(a_metric, "__class__"): name = a_metric.__class__.__name__
            else: name = "unknown_loss_name"
            # try:  # EAFP vs LBYL: both are duck-typing styles as they ask for what object can do (whether by introspection or trial) as opposed to checking its type.
            #     path = a_metric.path  # works for subclasses Metrics
            # except AttributeError: path = a_metric.__name__  # works for functions.
            loss_dict[name] = []
            for a_prediction, a_y_test in zip(prediction[0], groun_truth[0]):
                if hasattr(a_metric, "reset_states"): a_metric.reset_states()
                loss = a_metric(y_pred=a_prediction[None], y_true=a_y_test[None])
                loss_dict[name].append(np.array(loss).item())
        return pd.DataFrame(loss_dict)

    def save_class(self, weights_only: bool = True, version: str = '0', strict: bool = True, **kwargs: Any):
        """Simply saves everything:
        1. Hparams
        2. Data specs
        3. Model architecture or weights depending on the following argument.
        :param version: Model version, up to the user.
        :param weights_only: self-explanatory
        :return:
        """
        self.hp.save()  # goes into the meta path.
        self.data.save()  # goes into the meta path.
        tb.Save.vanilla_pickle(obj=self.history, path=self.hp.save_dir / 'metadata/training/history.pkl', verbose=True, desc="Training History")  # goes into the meta path.
        try: tb.Experimental.generate_readme(self.hp.save_dir, obj=self.__class__, **kwargs)
        except Exception as ex: print(ex)  # often fails because model is defined in main during experiments.
        save_dir = self.hp.save_dir.joinpath(f'{"weights" if weights_only else "model"}_save_v{version}').create()  # model save goes into data path.
        if weights_only: self.save_weights(save_dir)
        else: self.save_model(save_dir)

        import importlib
        __module = self.__class__.__module__
        if __module.startswith('__main__'):
            if strict: raise RuntimeError("Model class is defined in main. Saving the code from the current working directory. Consider importing the model class from a module.")
            else: pass
        try:
            module = importlib.import_module(__module)
        except ModuleNotFoundError as ex:
            print(ex)
            module = None
        if module is not None and hasattr(module, '__file__') and module.__file__ is not None:
            module_path_rh = tb.P(module.__file__).resolve().collapseuser().as_posix()
        else:
            module_path_rh = None
        specs = {'__module__': __module,
                 'model_class': self.__class__.__name__,
                 'data_class': self.data.__class__.__name__,
                 'hp_class': self.hp.__class__.__name__,
                 # the above is sufficient if module comes from installed package. Otherwise, if its from a repo, we need to add the following:
                 'module_path_rh': module_path_rh,
                 'cwd_rh': tb.P.cwd().collapseuser().as_posix(),
                 }
        tb.Save.json(obj=specs, path=self.hp.save_dir.joinpath('metadata/code_specs.json').str)
        print(f'SAVED Model Class @ {self.hp.save_dir.as_uri()}')
        return self.hp.save_dir

    @classmethod
    def from_class_weights(cls, path: Union[str, Path, P], hparam_class: Optional[Type[SubclassedHParams]] = None, data_class: Optional[Type[SubclassedDataReader]] = None,
                           device_name: Optional[Device] = None, verbose: bool = True):
        path = tb.P(path)
        if hparam_class is not None:
            hp_obj: SubclassedHParams = hparam_class.from_saved_data(path)
        else:
            hp_obj = tb.Read.vanilla_pickle(path=path / HParams.subpath + "hparams.HParams.pkl")
        if device_name: hp_obj.device_name = device_name
        if hp_obj.root != path.parent:
            hp_obj.root, hp_obj.name = path.parent, path.name  # if user moved the file to somewhere else, this will help alighment with new directory in case a modified version is to be saved.

        if data_class is not None: d_obj: SubclassedDataReader = data_class.from_saved_data(path, hp=hp_obj)
        else:
            d_obj = tb.Read.vanilla_pickle(path=path / DataReader.subpath / "data_reader.DataReader.pkl")
        # if type(hp_obj) is Generic[HParams]:
        d_obj.hp = hp_obj  # type: ignore
        # else:rd
            # raise ValueError(f"hp_obj must be of type `HParams` or `Generic[HParams]`. Got {type(hp_obj)}")
        model_obj = cls(hp_obj, d_obj)
        model_obj.load_weights(list(path.search('*_save_*'))[0])
        history_path = path / "metadata/training/history.pkl"
        if history_path.exists(): history: list[dict[str, Any]] = tb.Read.vanilla_pickle(path=history_path)
        else: history = []
        model_obj.history = history
        _ = print(f"LOADED {model_obj.__class__}: {model_obj.hp.name}") if verbose else None
        return model_obj

    @classmethod
    def from_class_model(cls, path: Union[str, Path, P]):
        path = tb.P(path)
        hp_obj = HParams.from_saved_data(path)
        data_obj = DataReader.from_saved_data(path, hp=hp_obj)
        directory = path.search('*_save_*')
        model_obj = cls.load_model(list(directory)[0])
        wrapper_class = cls(hp_obj, data_obj, model_obj)
        return wrapper_class

    @classmethod
    def from_path(cls, path_model: Union[str, Path, P], **kwargs: Any) -> 'SubclassedBaseModel':  # type: ignore
        path_model = tb.P(path_model).expanduser().absolute()
        specs = tb.Read.json(path=path_model.joinpath('metadata/code_specs.json'))
        print(f"Loading up module: `{specs['__module__']}`.")
        import importlib
        try:
            module = importlib.import_module(specs['__module__'])
        except ModuleNotFoundError as ex:
            print(ex)
            print(f"ModuleNotFoundError: Attempting to try again after appending path with `cwd`: `{specs['cwd_rh']}`.")
            import sys
            sys.path.append(tb.P(specs['cwd_rh']).expanduser().absolute().str)
            try:
                module = importlib.import_module(specs['__module__'])
            except ModuleNotFoundError as ex2:
                print(ex2)
                print(f"ModuleNotFoundError: Attempting to directly loading up `module_path`: `{specs['module_path_rh']}`.")
                module = load_class(tb.P(specs['module_path_rh']).expanduser().absolute().as_posix())
        model_class: SubclassedBaseModel = getattr(module, specs['model_class'])
        data_class: Type[DataReader] = getattr(module, specs['data_class'])
        hp_class: Type[HParams] = getattr(module, specs['hp_class'])
        return model_class.from_class_weights(path_model, hparam_class=hp_class, data_class=data_class, **kwargs)

    def plot_model(self, dpi: int = 150, **kwargs: Any):  # alternative viz via tf2onnx then Netron.
        import tensorflow as tf
        path = self.hp.save_dir.joinpath("metadata/model/model_plot.png")
        tf.keras.utils.plot_model(self.model, to_file=str(path), show_shapes=True, show_layer_names=True, show_layer_activations=True, show_dtype=True, expand_nested=True, dpi=dpi, **kwargs)
        print(f"Successfully plotted the model @ {path.as_uri()}")
        return path

    def build(self, sample_dataset: bool = False, ip_shapes: Optional[list[tuple[int, ...]]] = None, ip: Optional[list['npt.NDArray[np.float64]']] = None, verbose: bool = True):
        """ Building has two main uses.
        * Useful to baptize the model, especially when its layers are built lazily. Although this will eventually happen as the first batch goes in. This is a must before showing the summary of the model.
        * Doing sanity check about shapes when designing model.
        * Sanity check about values and ranges when random normal input is fed.
        :param sample_dataset:
        :param ip_shapes:
        :param ip:
        :param verbose:
        :return:
        """
        try:
            keys_ip = self.data.specs.get_split_strings(self.data.specs.ip_names, which_split="test")
            keys_op = self.data.specs.get_split_strings(self.data.specs.op_names, which_split="test")
        except TypeError as te:
            raise ValueError(f"Failed to load up sample data. Make sure that data has been loaded up properly.") from te

        if ip is None:
            if sample_dataset:
                ip, _, _ = self.data.sample_dataset()
            else:
                ip, _ = self.data.get_random_inputs_outputs(ip_shapes=ip_shapes)
        op = self.model(ip[0] if len(self.data.specs.ip_names) == 1 else ip)
        if not isinstance(op, list): ops = [op]
        else: ops = op
        ips = ip
        # ops = [op] if len(keys_op) == 1 else op
        # ips = [ip] if len(keys_ip) == 1 else ip
        if verbose:
            print("\n")
            print("Build Test".center(50, '-'))
            tb.Struct.from_keys_values(keys_ip, tb.L(ips).apply(lambda x: x.shape)).print(as_config=True, title="Input shapes:")
            tb.Struct.from_keys_values(keys_op, tb.L(ops).apply(lambda x: x.shape)).print(as_config=True, title=f"Output shape:")
            print("\n\nStats on output data for random normal input:")
            try:
                res = []
                for item_str, item_val in zip(keys_ip + keys_op, list(ips) + list(ops)):
                    a_df = pd.DataFrame(np.array(item_val).flatten()).describe().rename(columns={0: item_str})
                    res.append(a_df)
                print(pd.concat(res, axis=1))
            except Exception as ex:
                print(f"Could not do stats on outputs and inputs. Error: {ex}")
            print("Build Test Finished".center(50, '-'))
            print("\n")


class Ensemble(tb.Base):
    def __init__(self, hp_class: Type[SubclassedHParams], data_class: Type[SubclassedDataReader], model_class: Type[SubclassedBaseModel], size: int = 10, **kwargs: Any):
        """
        :param model_class: Either a class for constructing saved_models or list of saved_models already cosntructed.
          * In either case, the following methods should be implemented:
          __init__, load, load_weights, save, save_weights, predict, fit
          Model constructor should takes everything it needs from self.hp and self.data only.
          Otherwise, you must pass a list of already constructed saved_models.
        :param size: size of ensemble
        """
        super().__init__(**kwargs)
        self.__dict__.update(kwargs)
        self.size = size
        self.hp_class = hp_class
        self.data_class = data_class
        self.model_class = model_class
        self.models: list[BaseModel] = []
        # self.data = None  # one data object for all models (so that it can fit in the memory)
        if hp_class and data_class and model_class:
            # only generate the dataset once and attach it to the ensemble to be reused by models.
            self.data = self.data_class(hp=hp_class())  # type: ignore
            print("Creating Models".center(100, "="))
            for i in tqdm(range(size)):
                hp = self.hp_class()  # type: ignore
                hp.name = str(hp.name) + f'__model__{i}'
                datacopy: SubclassedDataReader = copy.copy(self.data)  # shallow copy
                datacopy.hp = hp  # type: ignore
                self.models.append(model_class(hp, datacopy))
        self.performance: list[Any] = []

    @classmethod
    def from_saved_models(cls, parent_dir: Union[str, Path, P], model_class: Type[SubclassedBaseModel], hp_class: Type[SubclassedHParams], data_class: Type[SubclassedDataReader]) -> 'Ensemble':
        obj = cls(hp_class=hp_class, data_class=data_class, model_class=model_class,  # type: ignore
                        path=parent_dir, size=len(tb.P(parent_dir).search('*__model__*')))
        obj.models = list(tb.P(parent_dir).search(pattern='*__model__*').apply(model_class.from_class_model))
        return obj

    @classmethod
    def from_saved_weights(cls, parent_dir: Union[str, Path, P], model_class: Type[SubclassedBaseModel], hp_class: Type[SubclassedHParams], data_class: Type[SubclassedDataReader]) -> 'Ensemble':
        obj = cls(model_class=model_class, hp_class=hp_class, data_class=data_class,  # type: ignore
                        path=parent_dir, size=len(tb.P(parent_dir).search('*__model__*')))
        obj.models = list(tb.P(parent_dir).search('*__model__*').apply(model_class.from_class_weights))  # type: ignore
        return obj

    @staticmethod
    def from_path(path: Union[str, Path, P]) -> list[SubclassedBaseModel]:  # type: ignore
        tmp = tb.P(path).expanduser().absolute().search("*")
        tmp2 = tmp.apply(BaseModel.from_path)
        return list(tmp2)  # type: ignore

    def fit(self, data_dict: dict[str, Any], populate_shapes: bool, shuffle_train_test: bool = True, save: bool = True, **kwargs: Any):
        self.performance = []
        for i in range(self.size):
            print('\n\n', f" Training Model {i} ".center(100, "*"), '\n\n')
            if shuffle_train_test:
                self.models[i].hp.seed = np.random.randint(0, 1000)
                self.data.split_the_data(data_dict=data_dict, populate_shapes=populate_shapes)  # shuffle data (shared among models)
            self.models[i].fit(**kwargs)
            self.performance.append(self.models[i].evaluate(aslice=slice(0, -1), viz=False))
            if save:
                self.models[i].save_class()
                tb.Save.vanilla_pickle(obj=self.performance, path=self.models[i].hp.save_dir / "performance.pkl")
        print("\n\n", f" Finished fitting the ensemble ".center(100, ">"), "\n")

    def clear_memory(self): pass  # t.cuda.empty_cache()


class Losses:
    @staticmethod
    def get_log_square_loss_class():
        import tensorflow as tf

        class LogSquareLoss(tf.keras.losses.Loss):
            def __init__(self, *args: Any, **kwargs: Any):
                super().__init__(*args, **kwargs)
                self.name = "LogSquareLoss"

            def call(self, y_true: 'npt.NDArray[np.float64]', y_pred: 'npt.NDArray[np.float64]'):
                _ = self
                tmp = tf.math.log(tf.convert_to_tensor(10.0, dtype=y_pred.dtype))  # type: ignore
                factor = tf.Tensor(20) / tmp  # type: ignore
                return factor * tf.math.log(tf.reduce_mean((y_true - y_pred)**2))
        return LogSquareLoss

    @staticmethod
    def get_mean_max_error(tf: Any):
        """
        For Tensorflow
        """
        class MeanMaxError(tf.keras.metrics.Metric):
            def __init__(self, name: str = 'MeanMaximumError', **kwargs: Any):
                super(MeanMaxError, self).__init__(name=name, **kwargs)
                self.mme = self.add_weight(name='mme', initializer='zeros')
                self.__name__ = name

            def update_state(self, y_true: 'npt.NDArray[np.float64]', y_pred: 'npt.NDArray[np.float64]', sample_weight: Optional['npt.NDArray[np.float64]'] = None): self.mme.assign(tf.reduce_mean(tf.reduce_max(sample_weight or 1.0 * tf.abs(y_pred - y_true), axis=1)))
            def result(self): return self.mme
            def reset_states(self): self.mme.assign(0.0)
        return MeanMaxError


class HPTuning:
    def __init__(self):
        # ================== Tuning ===============
        from tensorboard.plugins.hparams import api as hpt
        self.hpt = hpt
        import tensorflow as tf
        self.pkg = tf
        self.dir = None
        self.params = tb.List()
        self.acc_metric = None
        self.metrics = None

    @staticmethod
    def help():
        """Steps of use: subclass this and do the following:
        * Set directory attribute.
        * set params
        * set accuracy metric
        * generate writer.
        * implement run method.
        * run loop method.
        * in the command line, run `tensorboard --logdir <self.dir>`
        """
        pass

    def run(self, param_dict: dict[str, Any]):
        _, _ = self, param_dict
        # should return a result that you want to maximize
        return _

    # def gen_writer(self):
    #     import tensorflow as tf
    #     with tf.summary.create_file_writer(str(self.dir)).as_default():
    #         self.hpt.hparams_config(
    #             hparams=self.params,
    #             metrics=self.metrics)

    # def loop(self):
    #     import itertools
    #     counter = -1
    #     tmp = self.params.list[0].domain.values
    #     for combination in itertools.product(*[tmp]):
    #         counter += 1
    #         param_dict = dict(zip(self.params.list, combination))
    #         with self.pkg.summary.create_file_writer(str(self.dir / f"run_{counter}")).as_default():
    #             self.hpt.hparams(param_dict)  # record the values used in this trial
    #             accuracy = self.run(param_dict)
    #             self.pkg.summary.scalar(self.acc_metric, accuracy, step=1)

    # def optimize(self): self.gen_writer(); self.loop()


# class KerasOptimizer:
#     def __init__(self, d):
#         self.data = d
#         self.tuner = None

#     def __call__(self, ktp): pass

#     def tune(self):
#         kt = tb.install_n_import("kerastuner")
#         self.tuner = kt.Hyperband(self, objective='loss', max_epochs=10, factor=3, directory=tb.P.tmp('my_dir'), project_name='intro_to_kt')


def batcher(func_type: str = 'function'):
    if func_type == 'method':
        def batch(func: Callable[..., Any]):
            # from functools import wraps
            # @wraps(func)
            def wrapper(self: Any, x: Any, *args: Any, per_instance_kwargs: Optional[dict[str, Any]] = None, **kwargs: Any):
                output = []
                for counter, item in enumerate(x):
                    mykwargs = {key: value[counter] for key, value in per_instance_kwargs.items()} if per_instance_kwargs is not None else {}
                    output.append(func(self, item, *args, **mykwargs, **kwargs))
                return np.array(output)
            return wrapper
        return batch
    elif func_type == 'class': raise NotImplementedError
    elif func_type == 'function':
        class Batch(object):
            def __init__(self, func: Callable[[Any], Any]): self.func = func
            def __call__(self, x: Any, **kwargs: Any): return np.array([self.func(item, **kwargs) for item in x])
        return Batch


def batcherv2(func_type: str = 'function', order: int = 1):
    if func_type == 'method':
        def batch(func: Callable[[Any], Any]):
            # from functools import wraps
            # @wraps(func)
            def wrapper(self: Any, *args: Any, **kwargs: Any): return np.array([func(self, *items, *args[order:], **kwargs) for items in zip(*args[:order])])
            return wrapper
        return batch
    elif func_type == 'class': raise NotImplementedError
    elif func_type == 'function':
        class Batch(object):
            def __init__(self, func: Callable[[Any], Any]): self.func = func
            def __call__(self, *args: Any, **kwargs: Any): return np.array([self.func(self, *items, *args[order:], **kwargs) for items in zip(*args[:order])])
        return Batch


def load_class(file_path: str):
    import importlib.util
    module_spec = importlib.util.spec_from_file_location(name="__temp_module__", location=file_path)
    if module_spec is None: raise ValueError(f"Failed to load up module from path: {file_path}")
    module = importlib.util.module_from_spec(module_spec)
    assert module_spec.loader is not None, "Module loader is None."
    module_spec.loader.exec_module(module)
    return module


if __name__ == '__main__':
    pass




"""
Distributed Computing
"""


from typing import Optional, Any, Callable, Union
from math import ceil, floor
from enum import Enum
from dataclasses import dataclass
import psutil
import numpy as np
import crocodile.toolbox as tb
from crocodile.cluster.remote_machine import RemoteMachine, RemoteMachineConfig, WorkloadParams, LAUNCH_METHOD
from rich.console import Console
# from platform import system
# import time
# from rich.progress import track


console = Console()


class LoadCriterion(Enum):
    cpu = "cpu"
    ram = "ram"
    product = "cpu * ram"
    cpu_norm = "cpu_norm"
    ram_norm = "ram_norm"
    product_norm = "cpu_norm * ram_norm"


@dataclass
class MachineSpecs:
    cpu: float
    ram: float
    product: float
    cpu_norm: float
    ram_norm: float
    product_norm: float
    @staticmethod
    def get_this_machine_specs():
        cpu, ram = psutil.cpu_count(), psutil.virtual_memory().total / 2 ** 30
        return MachineSpecs(cpu=cpu, ram=ram, product=cpu * ram, cpu_norm=cpu, ram_norm=ram, product_norm=cpu * ram)


class ThreadLoadCalculator:
    """relies on relative values to a referenc machine specs.
    Runs multiple instances of code per machine. Useful if code doesn't run faster with more resources avaliable.
    equal distribution across instances of one machine"""
    def __init__(self, num_jobs: Optional[int] = None, load_criterion: LoadCriterion = LoadCriterion.cpu, reference_specs: Optional[MachineSpecs] = None):
        self.num_jobs = num_jobs
        self.load_criterion = load_criterion
        self.reference_specs: MachineSpecs = MachineSpecs.get_this_machine_specs() if reference_specs is None else reference_specs
    def __getstate__(self): return self.__dict__
    def __setstate__(self, state: dict[str, Any]): self.__dict__.update(state)
    def get_num_threads(self, machine_specs: MachineSpecs) -> int:
        if self.num_jobs is None: return 1
        res = int(floor(self.num_jobs * (machine_specs.__dict__[self.load_criterion.name] / self.reference_specs.__dict__[self.load_criterion.name])))
        return 1 if res == 0 else res


class MachineLoadCalculator:
    def __init__(self, max_num: int = 1000, load_criterion: LoadCriterion = LoadCriterion.product, load_ratios_repr: str = ""):
        self.load_ratios: list[float] = []
        self.load_ratios_repr = load_ratios_repr
        self.max_num: int = max_num
        self.load_criterion = load_criterion
    def __getstate__(self) -> dict[str, Any]: return self.__dict__
    def __setstate__(self, d: dict[str, Any]) -> None: self.__dict__.update(d)
    def get_workload_params(self, machines_specs: list[MachineSpecs], threads_per_machine: list[int]) -> list[WorkloadParams]:
        """Note: like thread divider in parallelize function, the behaviour is to include the edge cases on both ends of subsequent intervals."""
        tmp: list[WorkloadParams] = []
        idx_so_far = 0
        for machine_index, (machine_specs, a_threads_per_machine) in enumerate(zip(machines_specs, threads_per_machine)):
            load_value = machine_specs.__dict__[self.load_criterion.name]
            self.load_ratios.append(load_value)
            idx1 = idx_so_far
            idx2 = self.max_num if machine_index == len(threads_per_machine) - 1 else (floor(load_value * self.max_num) + idx1)
            if idx2 > self.max_num:
                print(machines_specs, '\n\n', threads_per_machine)
                print(f"All values: {tmp=}, {idx_so_far=}, {idx1=}, {idx2=}, {self.max_num=}, {a_threads_per_machine=}, {machine_index=}, {machine_specs=}, {load_value=}, {self.load_ratios=}, {self.load_ratios_repr=}")
                raise ValueError(f"idx2 ({idx2}) > max_num ({self.max_num})")
            idx_so_far = idx2
            tmp.append(WorkloadParams(idx_start=idx1, idx_end=idx2, idx_max=self.max_num, jobs=a_threads_per_machine))
        return tmp


class Cluster:
    def __getstate__(self) -> dict[str, Any]:
        state = self.__dict__
        state["func"] = None
        return state
    def __setstate__(self, state: dict[str, Any]) -> None: self.__dict__.update(state)
    def save(self) -> tb.P:
        path = self.root_dir.joinpath("cluster.Cluster.pkl")
        tb.Save.vanilla_pickle(obj=self.__getstate__(), path=path)
        return path
    @staticmethod
    def load(job_id: str, base: Optional[str] = None) -> 'Cluster': return Cluster.get_cluster_path(job_id=job_id, base=base).joinpath("cluster.Cluster.pkl").readit()
    @staticmethod
    def get_cluster_path(job_id: str, base: Union[str, tb.P, None] = None):
        if base is None: base_obj = tb.P.home().joinpath(rf"tmp_results/remote_machines")
        else: base_obj = tb.P(base)
        return base_obj.joinpath(f"job_id__{job_id}")
    def __init__(self,
                 func: Callable[..., Any],
                 ssh_params: list[dict[str, Any]],
                 remote_machine_config: RemoteMachineConfig,
                 func_kwargs: Optional[dict[str, Any]] = None,
                 # workload_params: list[WorkloadParams] or None = None,
                 thread_load_calc: Optional[ThreadLoadCalculator] = None,
                 # machine_load_calc=None,
                 ditch_unavailable_machines: bool = False,
                 description: str = "",
                 job_id: Optional[str] = None,
                 base_dir: Union[str, tb.P, None] = None):
        self.job_id = job_id or tb.randstr(noun=True)
        self.root_dir = self.get_cluster_path(self.job_id, base=base_dir)
        self.results_downloaded = False

        self.thread_load_calc: ThreadLoadCalculator = thread_load_calc or ThreadLoadCalculator()
        self.machine_load_calc: MachineLoadCalculator = MachineLoadCalculator(load_criterion=LoadCriterion[self.thread_load_calc.load_criterion.name + "_norm"], )

        sshz: list[tb.SSH] = []
        for an_ssh_params in ssh_params:
            try:
                tmp = tb.SSH(**an_ssh_params)
                sshz.append(tmp)
            except Exception as ex:
                print(f"Couldn't connect to {an_ssh_params}")
                if ditch_unavailable_machines: continue
                else: raise Exception(f"Couldn't connect to {an_ssh_params}") from ex  # type: ignore # pylint: disable=W0719

        # lists of similar length:
        self.sshz: list[tb.SSH] = sshz
        self.machines: list[RemoteMachine] = []
        self.machines_specs: list[MachineSpecs] = []
        self.threads_per_machine: list[int] = []
        self.remote_machine_kwargs: RemoteMachineConfig = remote_machine_config
        self.workload_params: list[WorkloadParams] = []

        self.description: str = description
        self.func = func
        self.func_kwargs = func_kwargs if func_kwargs is not None else {}

        # fire options
        self.machines_per_tab: int = 1
        self.window_number: int = 2

    def __repr__(self): return f"Cluster with following machines:\n" + "\n".join([repr(item) for item in (self.machines if self.machines else self.sshz)])
    def print_func_kwargs(self):
        print("\n" * 2)
        console.rule(title=f"kwargs of functions to be run on machines")
        for an_ssh, a_kwarg in zip(self.sshz, self.workload_params):
            tb.S(a_kwarg.__dict__).print(as_config=True, title=an_ssh.get_remote_repr())
    def print_commands(self, launch_method: LAUNCH_METHOD):
        print("\n" * 2)
        console.rule(title="Commands to run on each machine:")
        for machine in self.machines:
            print(f"{repr(machine)} ==> {machine.file_manager.get_fire_command(launch_method=launch_method)}")

    def generate_standard_kwargs(self) -> None:
        if self.workload_params:
            self.print_func_kwargs()
            print(self.workload_params, len(self.workload_params), type(self.workload_params))
            print("workload_params is not None, so not generating standard kwargs")
            return None
        cpus: list[float] = []
        rams: list[float] = []
        for an_ssh in self.sshz:
            res = an_ssh.run_py("import psutil; print(psutil.cpu_count(), psutil.virtual_memory().total)", verbose=False).op
            try: cpus.append(int(res.split(' ')[0]))
            except ValueError as ve:
                print(f"Couldn't get cpu count from {an_ssh}")
                raise ValueError(f"Couldn't get cpu count from {an_ssh.get_remote_repr()}") from ve
            rams.append(ceil(int(res.split(' ')[1]) / 2 ** 30))
        total_cpu = np.array(cpus).sum()
        total_ram = np.array(rams).sum()
        total_product = (np.array(cpus) * np.array(rams)).sum()

        self.machines_specs = [MachineSpecs(cpu=a_cpu, ram=a_ram, product=a_cpu * a_ram, cpu_norm=a_cpu / total_cpu, ram_norm=a_ram / total_ram, product_norm=a_cpu * a_ram / total_product) for a_cpu, a_ram in zip(cpus, rams)]
        self.threads_per_machine = [self.thread_load_calc.get_num_threads(machine_specs=machine_specs) for machine_specs in self.machines_specs]
        self.workload_params = self.machine_load_calc.get_workload_params(machines_specs=self.machines_specs, threads_per_machine=self.threads_per_machine)
        self.print_func_kwargs()

    def viz_load_ratios(self) -> None:
        if not self.workload_params: raise RuntimeError("func_kwargs_list is None. You need to run generate_standard_kwargs() first.")
        plt = tb.install_n_import("plotext")
        names = tb.L(self.sshz).apply(lambda x: x.get_remote_repr(add_machine=True)).list

        plt.simple_multiple_bar(names, [[machine_specs.cpu for machine_specs in self.machines_specs], [machine_specs.ram for machine_specs in self.machines_specs]], title=f"Resources per machine", labels=["#cpu threads", "memory size"])
        plt.show()
        print("")
        plt.simple_bar(names, self.machine_load_calc.load_ratios, width=100, title=f"Load distribution for machines using criterion `{self.machine_load_calc.load_criterion}`")
        plt.show()

        tmp = tb.S(dict(zip(names, tb.L((np.array(self.machine_load_calc.load_ratios) * 100).round(1)).apply(lambda x: f"{int(x)}%")))).print(as_config=True, justify=75, return_str=True)
        assert isinstance(tmp, str)
        self.machine_load_calc.load_ratios_repr = tmp
        print(self.machine_load_calc.load_ratios_repr)
        # self.workload_params.
        print("\n")

    def submit(self) -> None:
        if not self.workload_params: raise RuntimeError("You need to generate standard kwargs first.")
        for idx, (a_workload_params, an_ssh) in enumerate(zip(self.workload_params, self.sshz)):
            desc = self.description + f"\nLoad Ratios on machines:\n{self.machine_load_calc.load_ratios_repr}"
            # if self.remote_machine_kwargs is not None:
            config = self.remote_machine_kwargs
            config.__dict__.update(dict(description=desc, job_id=self.job_id + f"_{idx}", base_dir=self.root_dir, workload_params=a_workload_params, ssh_obj=an_ssh))
            # else: config = RemoteMachineConfig(description=desc, job_id=self.job_id + f"_{idx}", base_dir=self.root_dir.as_posix(), workload_params=a_workload_params, ssh_obj=an_ssh)
            m = RemoteMachine(func=self.func, func_kwargs=self.func_kwargs, config=config)
            m.generate_scripts()
            m.submit()
            self.machines.append(m)
        try: self.save()
        except Exception as re:
            print(re)
            print("Couldn't pickle cluster object")
        # self.print_commands()

    def open_mux(self, machines_per_tab: int = 1, window_number: Optional[int] = None):
        self.machines_per_tab = machines_per_tab
        self.window_number = window_number if window_number is not None else 0  # tb.randstr(length=3, lower=False, upper=False)
        cmd = f"wt -w {self.window_number} "
        for idx, m in enumerate(self.machines):

            sub_cmd = m.get_session_manager().get_new_session_ssh_command(ssh=m.ssh, sess_name=m.job_params.session_name)
            if idx == 0: cmd += f""" new-tab --title '{str(m.ssh.hostname) + str(idx)}' pwsh -Command "{sub_cmd}" `;"""  # avoid new tabs despite being even index
            elif idx % self.machines_per_tab == 0: cmd += f""" new-tab --title {str(m.ssh.hostname) + str(idx)} pwsh -Command "{sub_cmd}" `;"""
            else: cmd += f""" split-pane --horizontal --size {1 / self.machines_per_tab} pwsh -Command "{sub_cmd}" `;"""

        print("Terminal launch command:\n", cmd)
        if cmd.endswith("`;"): cmd = cmd[:-2]
        tb.Terminal().run_async(*cmd.replace("`;", ";").split(" "))  # `; only for powershell, cmd is okay for ; as it is not a special character
        rm_last = self.machines[-1]
        rm_last.get_session_manager().asssert_session_started(ssh=rm_last.ssh, sess_name=rm_last.job_params.session_name)

    def fire(self, machines_per_tab: int = 1, window_number: Optional[int] = None, run: bool = False):
        self.open_mux(machines_per_tab=machines_per_tab, window_number=window_number)
        for m in self.machines:
            m.fire(run=run, open_console=False)

    def run(self, run: bool = False, machines_per_tab: int = 1, window_number: Optional[int] = None):
        self.generate_standard_kwargs()
        self.viz_load_ratios()
        print(self)
        self.submit()
        self.fire(run=run, machines_per_tab=machines_per_tab, window_number=window_number)
        self.save()
        return self

    def check_job_status(self) -> None: tb.L(self.machines).apply(lambda machine: machine.check_job_status())
    def download_results(self):
        if self.results_downloaded:
            print(f"All results downloaded to {self.root_dir} 🤗")
            return True
        for idx, a_m in enumerate(self.machines):
            _ = idx
            if a_m.results_path is None:
                print(f"Results are not ready for machine {a_m}.")
                print(f"Try to run `.check_job_status()` to check if the job is done and obtain results path.")
                continue
            # results_folder = tb.P(a_m.results_path).expanduser()
            if a_m.results_downloaded is False:
                print("\n")
                console.rule(f"Downloading results from {a_m}")
                print("\n")
                a_m.download_results(target=None)  # TODO another way of resolve multiple machines issue is to create a directory at downlaod_results time.
        if tb.L(self.machines).results_downloaded.to_numpy().sum() == len(self.machines):
            print(f"All results downloaded to {self.root_dir} 🤗")
            self.results_downloaded = True


if __name__ == '__main__':
    pass




"""
Core
"""

from pathlib import Path
from typing import Optional, Union, Generic, TypeVar, Type, List as ListType, Any, Iterator, Callable, Iterable, Hashable, Protocol
import datetime

_Slice = TypeVar('_Slice', bound='Slicable')
class Slicable(Protocol):
    def __getitem__(self: _Slice, i: slice) -> _Slice: ...


T = TypeVar('T')
T2 = TypeVar('T2')
T3 = TypeVar('T3')
PLike = Union[str, Path]


# ============================== Accessories ============================================
def validate_name(astring: str, replace: str = '_') -> str: return __import__("re").sub(r'[^-a-zA-Z0-9_.()]+', replace, str(astring))
def timestamp(fmt: Optional[str] = None, name: Optional[str] = None) -> str: return ((name + '_') if name is not None else '') + __import__("datetime").datetime.now().strftime(fmt or '%Y-%m-%d-%I-%M-%S-%p-%f')  # isoformat is not compatible with file naming convention, fmt here is.
def str2timedelta(shift: str) -> datetime.timedelta:  # Converts a human readable string like '1m' or '1d' to a timedate object. In essence, its gives a `2m` short for `pd.timedelta(minutes=2)`"""
    key, val = {"s": "seconds", "m": "minutes", "h": "hours", "d": "days", "w": "weeks", "M": "months", "y": "years"}[shift[-1]], float(shift[:-1])
    key, val = ("days", val * 30) if key == "months" else (("weeks", val * 52) if key == "years" else (key, val)); return __import__("datetime").timedelta(**{key: val})
def install_n_import(library: str, package: Optional[str] = None, fromlist: Optional[list[str]] = None):  # sometimes package name is different from import, e.g. skimage.
    try: return __import__(library, fromlist=fromlist if fromlist is not None else ())
    except (ImportError, ModuleNotFoundError): __import__("subprocess").check_call([__import__("sys").executable, "-m", "pip", "install", package or library]); return __import__(library, fromlist=fromlist if fromlist is not None else ())
def randstr(length: int = 10, lower: bool = True, upper: bool = True, digits: bool = True, punctuation: bool = False, safe: bool = False, noun: bool = False) -> str:
    if safe: return __import__("secrets").token_urlsafe(length)  # interannly, it uses: random.SystemRandom or os.urandom which is hardware-based, not pseudo
    if noun: return install_n_import("randomname").get_name()
    string = __import__("string"); return ''.join(__import__("random").choices((string.ascii_lowercase if lower else "") + (string.ascii_uppercase if upper else "") + (string.digits if digits else "") + (string.punctuation if punctuation else ""), k=length))


def save_decorator(ext: str = ""):  # apply default paths, add extension to path, print the saved file path
    def decorator(func: Callable[..., Any]):
        def wrapper(obj: Any, path: Union[str, Path, None] = None, verbose: bool = True, add_suffix: bool = False, desc: str = "", class_name: str = "", **kwargs: Any):
            if path is None:
                path = Path.home().joinpath("tmp_results/tmp_files").joinpath(randstr(noun=True))
                _ = print(f"tb.core: Warning: Path not passed to {func}. A default path has been chosen: {Path(path).absolute().as_uri()}") if verbose else None
            if add_suffix:
                _ = [(print(f"tb.core: Warning: suffix `{a_suffix}` is added to path passed {path}") if verbose else None) for a_suffix in [ext, class_name] if a_suffix not in str(path)]
                path = str(path).replace(ext, "").replace(class_name, "") + class_name + ext
                path = Path(path).expanduser().resolve()
                path.parent.mkdir(parents=True, exist_ok=True)
            else: path = Path(path).expanduser().resolve()
            path.parent.mkdir(parents=True, exist_ok=True)
            func(path=path, obj=obj, **kwargs)
            if verbose: print(f"SAVED {desc or path.name} {obj.__class__.__name__}: {f(repr(obj), justify=0, limit=50)}  @ `{path.absolute().as_uri()}`. Size = {path.stat().st_size / 1024**2:0.2f} MB")  # |  Directory: `{path.parent.absolute().as_uri()}`
            return path
        return wrapper
    return decorator


@save_decorator(".json")
def json(obj: Any, path: PLike, indent: Optional[str] = None, encoding: str = 'utf-8', **kwargs: Any):
    _ = encoding
    return Path(path).write_text(__import__("json").dumps(obj, indent=indent, default=lambda x: x.__dict__, **kwargs), encoding="utf-8")
@save_decorator(".yml")
def yaml(obj: dict[Any, Any], path: PLike, **kwargs: Any):
    with open(Path(path), 'w', encoding="utf-8") as file: __import__("yaml").dump(obj, file, **kwargs)
@save_decorator(".toml")
def toml(obj: dict[Any, Any], path: PLike, encoding: str = 'utf-8'): return Path(path).write_text(install_n_import("toml").dumps(obj), encoding=encoding)
@save_decorator(".ini")
def ini(obj: dict[Any, Any], path: PLike, **kwargs: Any):
    conf = install_n_import("configparser").ConfigParser(); conf.read_dict(obj)
    with open(path, 'w', encoding="utf-8") as configfile: conf.write(configfile, **kwargs)
@save_decorator(".csv")
def csv(obj: Any, path: PLike): return obj.to_frame('dtypes').reset_index().to_csv(str(path) + ".dtypes")
@save_decorator(".npy")
def npy(obj: Any, path: PLike, **kwargs: Any): return __import__('numpy').save(path, obj, **kwargs)
# @save_decorator(".mat")
# def mat(mdict, path=None, **kwargs): _ = [mdict.__setitem(key, []) for key, value in mdict.items() if value is None]; from scipy.io import savemat; savemat(str(path), mdict, **kwargs)  # Avoid using mat as it lacks perfect restoration: * `None` type is not accepted. Scalars are conveteed to [1 x 1] arrays.
@save_decorator(".pkl")
def vanilla_pickle(obj: Any, path: PLike, **kwargs: Any): return Path(path).write_bytes(__import__("pickle").dumps(obj, **kwargs))
@save_decorator(".pkl")
def pickle(obj: Any, path: PLike, r: bool = False, **kwargs: Any): return Path(path).write_bytes(__import__("dill").dumps(obj, recurse=r, **kwargs))  # In IPyconsole of Pycharm, this works only if object is of a an imported class. Don't use with objects defined at main.
def pickles(obj: Any, r: bool = False, **kwargs: Any): return __import__("dill").dumps(obj, r=r, **kwargs)
class Save:
    json = json
    yaml = yaml
    toml = toml
    ini = ini
    csv = csv
    npy = npy
    # mat = mat
    vanilla_pickle = vanilla_pickle
    pickle = pickle
    pickles = pickles


# ====================================== Object Management ====================================
class Base(object):
    def __init__(self, *args: Any, **kwargs: Any): _ = args, kwargs
    def __getstate__(self) -> dict[str, Any]: return self.__dict__.copy()
    def __setstate__(self, state: dict[str, Any]): self.__dict__.update(state)
    def __deepcopy__(self, *args: Any, **kwargs: Any):
        obj = self.__class__(*args, **kwargs)
        obj.__dict__.update(__import__("copy").deepcopy(self.__dict__))
        return obj
    def __copy__(self, *args: Any, **kwargs: Any): obj = self.__class__(*args, **kwargs); obj.__dict__.update(self.__dict__.copy()); return obj
    # def eval(self, string_, func=False, other=False): return string_ if type(string_) is not str else eval((("lambda x, y: " if other else "lambda x:") if not str(string_).startswith("lambda") and func else "") + string_ + (self if False else ''))
    # def exec(self, expr: str) -> 'Base': exec(expr); return self  # exec returns None.
    def save(self, path: Union[str, Path, None] = None, add_suffix: bool = True, save_code: bool = False, verbose: bool = True, data_only: bool = True, desc: str = ""):  # + (".dat" if data_only else "")
        saved_file = Save.pickle(obj=self.__getstate__() if data_only else self, path=path, verbose=verbose, add_suffix=add_suffix, class_name="." + self.__class__.__name__, desc=desc or (f"Data of {self.__class__}" if data_only else desc))
        _ = self.save_code(path=saved_file.parent.joinpath(saved_file.name + "_saved_code.py")) if save_code else None; return self
    @classmethod
    def from_saved_data(cls, path: PLike, *args: Any, **kwargs: Any): obj = cls(*args, **kwargs); obj.__setstate__(dict(__import__("dill").loads(Path(path).read_bytes()))); return obj
    def save_code(self, path: Union[str, Path]):
        if hasattr(module := __import__("inspect").getmodule(self), "__file__"): file = Path(module.__file__)
        else: raise FileNotFoundError(f"Attempted to save code from a script running in interactive session! module should be imported instead.")
        _ = Path(path).expanduser().write_text(encoding='utf-8', data=file.read_text(encoding='utf-8')); return Path(path) if type(path) is str else path  # path could be tb.P, better than Path
    def get_attributes(self, remove_base_attrs: bool = True, return_objects: bool = False, fields: bool = True, methods: bool = True):
        import inspect
        remove_vals = Base().get_attributes(remove_base_attrs=False) if remove_base_attrs else []
        attrs: List[Any] = List(dir(self)).filter(lambda x: '__' not in x and not x.startswith('_')).remove(values=remove_vals)
        attrs: List[Any] = attrs.filter(lambda x: (inspect.ismethod(getattr(self, x)) if not fields else True) and ((not inspect.ismethod(getattr(self, x))) if not methods else True))  # logic (questionable): anything that is not a method is a field
        return List([getattr(self, x) for x in attrs]) if return_objects else List(attrs)
    def print(self, dtype: bool = False, attrs: bool = False, **kwargs: Any): return Struct(self.__dict__).update(attrs=self.get_attributes() if attrs else None).print(dtype=dtype, **kwargs)
    @staticmethod
    def get_state(obj: Any, repr_func: Callable[[Any], dict[str, Any]] = lambda x: x, exclude: Optional[list[str]] = None) -> dict[str, Any]:
        if not any([hasattr(obj, "__getstate__"), hasattr(obj, "__dict__")]): return repr_func(obj)
        return (tmp if type(tmp := obj.__getstate__() if hasattr(obj, "__getstate__") else obj.__dict__) is not dict else Struct(tmp).filter(lambda k, v: k not in (exclude or [])).apply2values(lambda k, v: Base.get_state(v, exclude=exclude, repr_func=repr_func)).__dict__)
    def viz_composition_heirarchy(self, depth: int = 3, obj: Any = None, filt: Optional[Callable[[Any], None]] = None):
        install_n_import("objgraph").show_refs([self] if obj is None else [obj], max_depth=depth, filename=str(filename := Path(__import__("tempfile").gettempdir()).joinpath("graph_viz_" + randstr(noun=True) + ".png")), filter=filt)
        _ = __import__("os").startfile(str(filename.absolute())) if __import__("sys").platform == "win32" else None; return filename


class List(Generic[T]):  # Inheriting from Base gives save method.  # Use this class to keep items of the same type."""
    def __init__(self, obj_list: Union[ListType[T], None, Iterator[T], Iterable[T]] = None) -> None: super().__init__(); self.list = list(obj_list) if obj_list is not None else []
    def __repr__(self): return f"List [{len(self.list)} elements]. First Item: " + f"{get_repr(self.list[0], justify=0, limit=100)}" if len(self.list) > 0 else f"An Empty List []"
    def print(self, sep: str = '\n', styler: Callable[[Any], str] = repr, return_str: bool = False, **kwargs: dict[str, Any]):
        res = sep.join([f"{idx:2}- {styler(item)}" for idx, item in enumerate(self.list)])
        _ = print(res) if not return_str else None; _ = kwargs
        return res if return_str else None
    def __deepcopy__(self, arg: Any) -> "List[T]": _ = arg; return List([__import__("copy").deepcopy(i) for i in self.list])
    def __bool__(self) -> bool: return bool(self.list)
    def __contains__(self, key: str) -> bool: return key in self.list
    def __copy__(self) -> 'List[T]': return List(self.list.copy())
    def __getstate__(self) -> list[T]: return self.list
    def __setstate__(self, state: list[T]): self.list = state
    def __len__(self) -> int: return len(self.list)
    def __iter__(self) -> Iterator[T]: return iter(self.list)
    def __array__(self): import numpy as np; return np.array(self.list)  # compatibility with numpy
    # def __next__(self) -> T: return next(self.list)
    @property
    def len(self) -> int: return len(self.list)
    # ================= call methods =====================================
    def __getattr__(self, name: str) -> 'List[T]': return List(getattr(i, name) for i in self.list)  # fallback position when __getattribute__ mechanism fails.
    def __call__(self, *args: Any, **kwargs: Any) -> 'List[Any]':
        items = self.list
        return List([ii.__call__(*args, **kwargs) for ii in items])  # type: ignore
    # ======================== Access Methods ==========================================
    def __setitem__(self, key: int, value: T) -> None: self.list[key] = value
    def sample(self, size: int = 1, replace: bool = False, p: Optional[list[float]] = None) -> 'List[T]':
        import numpy as np
        tmp = np.random.choice(len(self), size, replace=replace, p=p)
        return List([self.list[item] for item in tmp.tolist()])
    def split(self, every: int = 1, to: Optional[int] = None) -> 'List[List[T]]':
        every = every if to is None else __import__("math").ceil(len(self) / to)
        res: list[List[T]] = []
        for ix in range(0, len(self), every):
            if ix + every < len(self):
                tmp = self.list[ix:ix + every]
            else:
                tmp = self.list[ix:len(self)]
            res.append(List(tmp))
        return List(res)
    def filter(self, func: Callable[[T], bool], which: Callable[[int, T], Union[T, T2]] = lambda _idx, _x: _x) -> 'List[Union[T2, T]]':
        return List([which(idx, x) for idx, x in enumerate(self.list) if func(x)])
    # ======================= Modify Methods ===============================
    def reduce(self, func: Callable[[T, T], T] = lambda x, y: x + y, default: Optional[T] = None) -> 'List[T]':  # type: ignore
        from functools import reduce
        if default is None: return List(reduce(func, self.list))  # type: ignore
        return List(reduce(func, self.list, default))  # type: ignore
    def append(self, item: T) -> 'List[T]': self.list.append(item); return self
    def __add__(self, other: 'List[T]') -> 'List[T]': return List(self.list + list(other))  # implement coersion
    def __radd__(self, other: 'List[T]') -> 'List[T]': return List(list(other) + self.list)
    def __iadd__(self, other: 'List[T]') -> 'List[T]': self.list = self.list + list(other); return self  # inplace add.
    def sort(self, key: Callable[[T], float], reverse: bool = False) -> 'List[T]': self.list.sort(key=key, reverse=reverse); return self
    def sorted(self, *args: list[Any], **kwargs: Any) -> 'List[T]': return List(sorted(self.list, *args, **kwargs))
    def insert(self, __index: int, __object: T): self.list.insert(__index, __object); return self
    # def modify(self, expr: str, other: Optional['List[T]'] = None) -> 'List[T]': _ = [exec(expr) for idx, x in enumerate(self.list)] if other is None else [exec(expr) for idx, (x, y) in enumerate(zip(self.list, other))]; return self
    def remove(self, value: Optional[T] = None, values: Optional[list[T]] = None, strict: bool = True) -> 'List[T]':
        for a_val in ((values or []) + ([value] if value else [])):
            if strict or value in self.list: self.list.remove(a_val)
        return self
    def to_series(self): return __import__("pandas").Series(self.list)
    def to_list(self) -> list[T]: return self.list
    def to_numpy(self, **kwargs: Any) -> 'Any': import numpy as np; return np.array(self.list, **kwargs)
    def to_struct(self, key_val: Optional[Callable[[T], Any]] = None) -> 'Struct': return Struct.from_keys_values_pairs(self.apply(func=key_val if key_val else lambda x: (str(x), x)))
    # def index(self, val: int) -> int: return self.list.index(val)
    def slice(self, start: Optional[int] = None, stop: Optional[int] = None, step: Optional[int] = None) -> 'List[T]': return List(self.list[start:stop:step])
    def __getitem__(self, key: Union[int, list[int], '_Slice']) -> Union[T, 'List[T]']:
        if isinstance(key, (list, Iterable, Iterator)): return List(self.list[item] for item in key)  # to allow fancy indexing like List[1, 5, 6]
        # elif isinstance(key, str): return List(item[key] for item in self.list)  # access keys like dictionaries.
        elif isinstance(key, int): return self.list[key]
        # assert isinstance(key, slice)
        return List(self.list[key])  # type: ignore # noqa: call-overload  # slices
    def apply(self, func: Union[Callable[[T], T2], Callable[[T, T2], T3]], *args: Any, other: Optional['List[T]'] = None, filt: Callable[[T], bool] = lambda _x: True,
              jobs: Optional[int] = None, prefer: Optional[str] = [None, 'processes', 'threads'][0], verbose: bool = False, desc: Optional[str] = None, **kwargs: Any,
              ) -> 'Union[List[Union[T2, T3]], List[T3]]':
        # if depth > 1: self.apply(lambda x: x.apply(func, *args, other=other, jobs=jobs, depth=depth - 1, **kwargs))
        from tqdm import tqdm
        iterator: Iterable[Any]
        if other is None:
            iterator = (self.list if not verbose else tqdm(self.list, desc=desc))
        else: iterator = (zip(self.list, other) if not verbose else tqdm(zip(self.list, other), desc=desc))
        if jobs:
            from joblib import Parallel, delayed
            if other is None: return List(Parallel(n_jobs=jobs, prefer=prefer)(delayed(func)(x, *args, **kwargs) for x in iterator))  # type: ignore
            return List(Parallel(n_jobs=jobs, prefer=prefer)(delayed(func)(x, y) for x, y in iterator))  # type: ignore
        if other is None:
            return List([func(x, *args, **kwargs) for x in iterator if filt(x)])
        # func_: Callable[[T, T2], T3]
        return List([func(x, y) for x, y in iterator])  # type: ignore
    def to_dataframe(self, names: Optional[list[str]] = None, minimal: bool = False, obj_included: bool = True):
        import pandas as pd
        df = pd.DataFrame(columns=(['object'] if obj_included or names else []) + list(self.list[0].__dict__.keys()))
        if minimal: return df
        for i, obj in enumerate(self.list):  # Populate the dataframe:
            if obj_included or names:
                tmp: list[Any] = list(self.list[i].__dict__.values())
                data: list[Any] = [obj] if names is None else [names[i]]
                df.iloc[i] = pd.Series(data + tmp)
            else: df.iloc[i] = pd.Series(self.list[i].__dict__.values())  # type: ignore
        return df


class Struct(Base):  # inheriting from dict gives `get` method, should give `__contains__` but not working. # Inheriting from Base gives `save` method.
    """Use this class to keep bits and sundry items. Combines the power of dot notation in classes with strings in dictionaries to provide Pandas-like experience"""
    def __init__(self, dictionary: Union[dict[Any, Any], Type[object], None] = None, **kwargs: Any):
        if dictionary is None or isinstance(dictionary, dict): final_dict: dict[str, Any] = {} if dictionary is None else dictionary
        else:
            final_dict = (dict(dictionary) if dictionary.__class__.__name__ == "mappingproxy" else dictionary.__dict__)  # type: ignore
        final_dict.update(kwargs)  # type ignore
        super(Struct, self).__init__(); self.__dict__ = final_dict  # type: ignore
    # @staticmethod
    # def recursive_struct(mydict: dict[Any, Any]) -> 'Struct': struct = Struct(mydict); [struct.__setitem__(key, Struct.recursive_struct(val) if type(val) is dict else val) for key, val in struct.items()]; return struct
    # @staticmethod
    # def recursive_dict(struct) -> 'Struct': _ = [struct.__dict__.__setitem__(key, Struct.recursive_dict(val) if type(val) is Struct else val) for key, val in struct.__dict__.items()]; return struct.__dict__
    def save_json(self, path: Optional[PLike] = None, indent: Optional[str] = None): return Save.json(obj=self.__dict__, path=path, indent=indent)
    @classmethod
    def from_keys_values(cls, k: Iterable[str], v: Iterable[Any]) -> 'Struct': return Struct(dict(zip(k, v)))
    from_keys_values_pairs = classmethod(lambda cls, my_list: cls({k: v for k, v in my_list}))
    @classmethod
    def from_names(cls, names: list[str], default_: Optional[Any] = None) -> 'Struct': return cls.from_keys_values(k=names, v=default_ or [None] * len(names))  # Mimick NamedTuple and defaultdict
    def spawn_from_values(self, values: Union[list[Any], List[Any]]) -> 'Struct': return self.from_keys_values(list(self.keys()), values)
    def spawn_from_keys(self, keys: Union[list[str], List[str]]) -> 'Struct': return self.from_keys_values(keys, list(self.values()))
    def to_default(self, default: Optional[Callable[[], Any]] = lambda: None): tmp2 = __import__("collections").defaultdict(default); tmp2.update(self.__dict__); self.__dict__ = tmp2; return self
    def __str__(self, sep: str = "\n"): return config(self.__dict__, sep=sep)
    def __getattr__(self, item: str) -> 'Struct':
        try: return self.__dict__[item]
        except KeyError as ke: raise AttributeError(f'{type(self).__name__!r} object has no attribute {item!r}') from ke  # this works better with the linter. replacing Key error with Attribute error makes class work nicely with hasattr() by returning False.
    clean_view = property(lambda self: type("TempClass", (object,), self.__dict__))
    def __repr__(self, limit: int = 150): return "Struct: " + Display.get_repr(self.keys().list.__repr__(), limit=limit, justify=0)
    def __getitem__(self, item: str): return self.__dict__[item]  # thus, gives both dot notation and string access to elements.
    def __setitem__(self, key: str, value: Any): self.__dict__[key] = value
    def __bool__(self): return bool(self.__dict__)
    def __contains__(self, key: Hashable): return key in self.__dict__
    def __len__(self): return len(self.keys())
    def __getstate__(self): return self.__dict__  # serialization
    def __setstate__(self, state: dict[Any, Any]): self.__dict__ = state
    def __iter__(self): return iter(self.__dict__.items())
    def __delitem__(self, key: str): del self.__dict__[key]
    def copy(self) -> 'Struct': return Struct(self.__dict__.copy())
    def to_dataframe(self, *args: Any, **kwargs: Any): return __import__("pandas").DataFrame(self.__dict__, *args, **kwargs)
    def keys(self, verbose: bool = False) -> 'List[Any]': return List(list(self.__dict__.keys())) if not verbose else install_n_import("tqdm").tqdm(self.__dict__.keys())
    def values(self, verbose: bool = False) -> 'List[Any]': return List(list(self.__dict__.values())) if not verbose else install_n_import("tqdm").tqdm(self.__dict__.values())
    def items(self, verbose: bool = False, desc: str = "") -> 'List[Any]': return List(self.__dict__.items()) if not verbose else install_n_import("tqdm").tqdm(self.__dict__.items(), desc=desc)
    def get(self, key: Optional[str] = None, default: Optional[Any] = None, strict: bool = False, keys: Union[None, list[str]] = None) -> 'Union[Any, List[Any]]':
        if keys is not None: return List([self.__dict__.get(key, default) if not strict else self[key] for key in keys])
        if key is not None: return (self.__dict__.get(key, default) if not strict else self[key])
        else: raise ValueError("Either key or keys should be passed.")
    def apply2keys(self, kv_func: Callable[[Any, Any], Any], verbose: bool = False, desc: str = "") -> 'Struct': return Struct({kv_func(key, val): val for key, val in self.items(verbose=verbose, desc=desc)})
    def apply2values(self, kv_func: Callable[[Any, Any], Any], verbose: bool = False, desc: str = "") -> 'Struct': _ = [self.__setitem__(key, kv_func(key, val)) for key, val in self.items(verbose=verbose, desc=desc)]; return self
    def apply(self, kv_func: Callable[[Any, Any], Any]) -> 'List[Any]': return self.items().apply(lambda item: kv_func(item[0], item[1]))
    def filter(self, kv_func: Callable[[Any, Any], Any]) -> 'Struct': return Struct({key: self[key] for key, val in self.items() if kv_func(key, val)})
    def inverse(self) -> 'Struct': return Struct({v: k for k, v in self.__dict__.items()})
    def update(self, *args: Any, **kwargs: Any) -> 'Struct': self.__dict__.update(Struct(*args, **kwargs).__dict__); return self
    def delete(self, key: Optional[str] = None, keys: Optional[list[str]] = None, kv_func: Optional[Callable[[Any, Any], Any]] = None) -> 'Struct':
        for key in ([key] if key else [] + (keys if keys is not None else [])): self.__dict__.__delitem__(key)
        if kv_func is not None:
            for k, v in self.items():
                if kv_func(k, v): self.__dict__.__delitem__(k)
        return self
    def _pandas_repr(self, justify: int, return_str: bool = False, limit: int = 30):
        import pandas as pd
        import numpy as np
        col2: List[Any] = self.values().apply(lambda x: str(type(x)).split("'")[1])
        col3: List[Any] = self.values().apply(lambda x: get_repr(x, justify=justify, limit=limit).replace("\n", " "))
        array = np.array([self.keys(), col2, col3]).T
        res: pd.DataFrame = pd.DataFrame(array, columns=["key", "dtype", "details"])
        return res if not return_str else str(res)
    def print(self, dtype: bool = True, return_str: bool = False, justify: int = 30, as_config: bool = False, as_yaml: bool = False,  # type: ignore # pylint: disable=W0237
              limit: int = 50, title: str = "", attrs: bool = False, **kwargs: Any) -> Union[str, None]:  # type: ignore
        _ = attrs
        import pandas as pd
        if as_config and not return_str:
            install_n_import("rich").inspect(self, value=False, title=title, docs=False, sort=False)
            return None
        if not bool(self):
            if return_str: return f"Empty Struct."
            else: print(f"Empty Struct."); return None
        else:
            if as_yaml or as_config:
                tmp: str = __import__("yaml").dump(self.__dict__) if as_yaml else config(self.__dict__, justify=justify, **kwargs)
                if return_str: return tmp
                else: print(tmp); return None
            else:
                tmp2 = self._pandas_repr(justify=justify, return_str=False, limit=limit)
                if isinstance(tmp2, pd.DataFrame):
                    res = tmp2.drop(columns=[] if dtype else ["dtype"])
                else: raise TypeError(f"Unexpected type {type(tmp2)}")
                if not return_str:
                    if install_n_import("tabulate"):
                        install_n_import("rich").print(res.to_markdown())
                    else: print(res)
                    return None
                return str(res)

    @staticmethod
    def concat_values(*dicts: dict[Any, Any], orient: str = 'list') -> 'Struct':
        import pandas as pd
        assert orient in ["dict", "list", "series", "split", "tight", "index"]
        return Struct(pd.concat(List(dicts).apply(lambda x: Struct(x).to_dataframe()).list).to_dict(orient=orient))  # type: ignore
    def plot_plt(self, title: str = '', xlabel: str = '', ylabel: str = '', **kwargs: Any):
        from crocodile.matplotlib_management import LineArtist
        artist = LineArtist(figname='Structure Plot', **kwargs)
        artist.plot_dict(self.__dict__, title=title, xlabel=xlabel, ylabel=ylabel)
        return artist
    def plot_plotly(self):
        from crocodile.plotly_management import px
        fig = px.line(self.__dict__)
        fig.show()
        return fig


def set_pandas_display(rows: int = 1000, columns: int = 1000, width: int = 5000, colwidth: int = 40) -> None:
    import pandas as pd; pd.set_option('display.max_colwidth', colwidth); pd.set_option('display.max_columns', columns); pd.set_option('display.width', width); pd.set_option('display.max_rows', rows)
def set_pandas_auto_width(): __import__("pandas").set_option('width', 0)  # this way, pandas is told to detect window length and act appropriately.  For fixed width host windows, this is recommended to avoid chaos due to line-wrapping.
def set_numpy_display(precision: int = 3, linewidth: int = 250, suppress: bool = True, floatmode: str = 'fixed', **kwargs: Any) -> None: __import__("numpy").set_printoptions(precision=precision, suppress=suppress, linewidth=linewidth, floatmode=floatmode, **kwargs)
def config(mydict: dict[Any, Any], sep: str = "\n", justify: int = 15, quotes: bool = False): return sep.join([f"{key:>{justify}} = {repr(val) if quotes else val}" for key, val in mydict.items()])
def f(str_: str, limit: int = 10000000000, justify: int = 50, direc: str = "<") -> str: return f"{(str_[:limit - 4] + '... ' if len(str_) > limit else str_):{direc}{justify}}"
def eng(): __import__("pandas").set_eng_float_format(accuracy=3, use_eng_prefix=True); __import__("pandas").options.float_format = '{:, .5f}'.format; __import__("pandas").set_option('precision', 7)  # __import__("pandas").set_printoptions(formatter={'float': '{: 0.3f}'.format})
def outline(array: 'Any', name: str = "Array", printit: bool = True): str_ = f"{name}. Shape={array.shape}. Dtype={array.dtype}"; _ = print(str_) if printit else None; return str_
def get_repr(data: Any, justify: int = 15, limit: int = 10000, direc: str = "<") -> str:
    if (dtype := data.__class__.__name__) in {'list', 'str'}: str_ = data if dtype == 'str' else f"list. length = {len(data)}. " + ("1st item type: " + str(type(data[0])).split("'")[1]) if len(data) > 0 else " "
    elif dtype in {"DataFrame", "Series"}: str_ = f"Pandas DF: shape = {data.shape}, dtype = {data.dtypes}." if dtype == 'DataFrame' else f"Pandas Series: Length = {len(data)}, Keys = {get_repr(data.keys().to_list())}."
    else: str_ = f"shape = {data.shape}, dtype = {data.dtype}." if dtype == 'ndarray' else repr(data)
    return f(str_.replace("\n", ", "), justify=justify, limit=limit, direc=direc)
def print_string_list(mylist: list[Any], char_per_row: int = 125, sep: str = " ", style: Callable[[Any], str] = str, _counter: int = 0):
    for item in mylist: _ = print("") if (_counter + len(style(item))) // char_per_row > 0 else print(style(item), end=sep); _counter = len(style(item)) if (_counter + len(style(item))) // char_per_row > 0 else _counter + len(style(item))
class Display:
    set_pandas_display = staticmethod(set_pandas_display)
    set_pandas_auto_width = staticmethod(set_pandas_auto_width)
    set_numpy_display = staticmethod(set_numpy_display)
    config = staticmethod(config)
    f = staticmethod(f)
    eng = staticmethod(eng)
    outline = staticmethod(outline)
    get_repr = staticmethod(get_repr)
    print_string_list = staticmethod(print_string_list)  # or D = type('D', (object, ), dict(set_pandas_display


if __name__ == '__main__':
    pass




# """
# References:
#     * https://learn.microsoft.com/en-us/graph/tutorials/python?tabs=aad
#     * https://github.com/pranabdas/Access-OneDrive-via-Microsoft-Graph-Python

# """

# import requests
# import json
# import urllib
# # import os
# from getpass import getpass
# import time
# from datetime import datetime
# import crocodile.toolbox as tb


# URL_AUTH = 'https://login.microsoftonline.com/common/oauth2/v2.0/authorize'
# URL_TOKEN = 'https://login.microsoftonline.com/common/oauth2/v2.0/token'
# URL_GRAPH = 'https://graph.microsoft.com/v1.0/'
# redirect_uri = 'http://localhost:8080/'


# class OneDrive:
#     def __init__(self, account=None):
#         self.creds_path = tb.P.home().joinpath("dotfiles/microsoft/onedrive/config.toml")
#         self.creds_content = self.creds_path.readit()
#         self.account = account
#         creds = self.creds_content[self.creds_content['accounts']['default']] if account is None else self.creds_content[account]
#         self.creds = creds
#         self.client_id = creds['client_id']
#         self.client_secret = creds['client_secret']
#         self.permissions = ['offline_access', 'files.readwrite', 'User.Read']

#         self.token, self.refresh_token, self.last_updated = creds['token'], creds['refresh_token'], creds['last_updated']
#         self.headers = {'Authorization': 'Bearer ' + self.token}

#         scope = ''
#         for items in range(len(self.permissions)):
#             scope = scope + self.permissions[items]
#             if items < len(self.permissions)-1:
#                 scope = scope + '+'
#         self.scope = scope
#         self.get_refresh_token()

#     def save_creds(self):
#         creds = self.creds
#         creds['token'] = self.token
#         creds['refresh_token'] = self.refresh_token
#         creds['last_updated'] = self.last_updated
#         self.creds_path.writeit(creds)

#     def token_flow_auth(self):
#         response_type = 'token'
#         print('Click over this link ' + URL_AUTH + '?client_id=' + self.client_id + '&scope=' + self.scope + '&response_type=' + response_type + '&redirect_uri=' + urllib.parse.quote(redirect_uri))
#         print('Sign in to your account, copy the whole redirected URL.')
#         code = input("Paste the URL here :")
#         token = code[(code.find('access_token') + len('access_token') + 1): (code.find('&token_type'))]
#         headers = {'Authorization': 'Bearer ' + token}
#         response = requests.get(URL_GRAPH + 'me/drive/', headers=headers)
#         print(json.loads(response.text))
#         if response.status_code == 200:
#             response = json.loads(response.text)
#             print('Connected to the OneDrive of', response['owner']['user']['displayName']+' (', response['driveType']+' ).', '\nConnection valid for one hour. Reauthenticate if required.')
#         elif response.status_code == 401:
#             response = json.loads(response.text)
#             print('API Error! : ', response['error']['code'], '\nSee response for more details.')
#         else:
#             response = json.loads(response.text)
#             print('Unknown error! See response for more details.')
#             print(response)

#     def code_flow_auth(self):
#         # returns access_token and refresh_token  for persistent session.
#         response_type = 'code'
#         print('Click over this link ' + URL_AUTH + '?client_id=' + self.client_id + '&scope=' + self.scope + '&response_type=' + response_type + '&redirect_uri=' + urllib.parse.quote(redirect_uri))
#         print('Sign in to your account, copy the whole redirected URL.')
#         code = getpass("Paste the URL here :")
#         code = code[(code.find('?code') + len('?code') + 1):]
#         response = requests.post(URL_TOKEN + '?client_id=' + self.client_id + '&scope=' + self.scope + '&grant_type=authorization_code' + '&redirect_uri=' + urllib.parse.quote(redirect_uri) + '&code=' + code)
#         _ = response
#         data = {
#             "client_id": self.client_id,
#             "scope": self.permissions,
#             "code": code,
#             "redirect_uri": redirect_uri,
#             "grant_type": 'authorization_code',
#             "client_secret": self.client_secret
#         }

#         response = requests.post(URL_TOKEN, data=data)
#         try:
#             token = json.loads(response.text)["access_token"]
#             refresh_token = json.loads(response.text)["refresh_token"]
#         except KeyError:
#             print(response.text)
#             raise KeyError
#         last_updated = time.mktime(datetime.today().timetuple())
#         self.token, self.refresh_token, self.last_updated = token, refresh_token, last_updated

#     def get_refresh_token(self):
#         data = {
#             "client_id": self.client_id,
#             "scope": self.permissions,
#             "refresh_token": self.refresh_token,
#             "redirect_uri": redirect_uri,
#             "grant_type": 'refresh_token',
#             "client_secret": self.client_secret,
#         }
#         response = requests.post(URL_TOKEN, data=data)
#         try:
#             token = json.loads(response.text)["access_token"]
#             refresh_token = json.loads(response.text)["refresh_token"]
#         except KeyError:
#             print(response.text)
#             raise KeyError
#         last_updated = time.mktime(datetime.today().timetuple())
#         self.token, self.refresh_token, self.last_updated = token, refresh_token, last_updated
#         self.headers = {'Authorization': 'Bearer ' + self.token}

#     def connect(self):
#         response = requests.get(URL_GRAPH + 'me/drive/', headers=self.headers)

#         if response.status_code == 200:
#             response = json.loads(response.text)
#             print('Connected to the OneDrive of', response['owner']['user']['displayName']+' (', response['driveType']+' ).', '\nConnection valid for one hour. Refresh token if required.')
#         elif response.status_code == 401:
#             response = json.loads(response.text)
#             print('API Error! : ', response['error']['code'], '\nSee response for more details.')
#             print(response)
#         else:
#             response = json.loads(response.text)
#             print('Unknown error! See response for more details.')

#     def list_items(self):
#         items = json.loads(requests.get(URL_GRAPH + 'me/drive/root/children', headers=self.headers).text)
#         items = items['value']


# def main():
#     pass


# if __name__ == '__main__':
#     d = OneDrive()




"""RM
"""

from typing import Optional, Any, Union, Callable
from dataclasses import dataclass, field
import time
import platform
import getpass
import crocodile.toolbox as tb
from crocodile.cluster.session_managers import Zellij, WindowsTerminal
from crocodile.cluster.self_ssh import SelfSSH
from crocodile.cluster.loader_runner import JobParams, EmailParams, WorkloadParams, FileManager, TRANSFER_METHOD, LAUNCH_METHOD, JOB_STATUS, CloudManager, LogEntry
import crocodile.cluster as cluster
from rich.panel import Panel
from rich.syntax import Syntax
from rich import inspect
# from rich.text import Text
from rich.console import Console
import pandas as pd


console = Console()


@dataclass
class RemoteMachineConfig:
    # conn
    job_id: str = field(default_factory=lambda: tb.randstr(noun=True))
    base_dir: str = f"~/tmp_results/remote_machines/jobs"
    description: str = ""
    ssh_params: dict[str, Union[str, int]] = field(default_factory=lambda: {})
    ssh_obj: Union[tb.SSH, SelfSSH, None] = None

    # data
    copy_repo: bool = False
    update_repo: bool = False
    install_repo: bool = False
    update_essential_repos: bool = True
    data: Optional[list[Any]] = None
    transfer_method: TRANSFER_METHOD = "sftp"
    cloud_name: Optional[str] = None

    # remote machine behaviour
    allowed_remotes: Optional[list[str]] = None
    open_console: bool = True
    notify_upon_completion: bool = False
    to_email: Optional[str] = None
    email_config_name: Optional[str] = None

    # execution behaviour
    launch_method: LAUNCH_METHOD = "remotely"
    kill_on_completion: bool = False
    ipython: bool = False
    interactive: bool = False
    pdb: bool = False
    pudb: bool = False
    wrap_in_try_except: bool = False
    parallelize: bool = False
    lock_resources: bool = True
    max_simulataneous_jobs: int = 1
    workload_params: Optional[WorkloadParams] = None
    def __post_init__(self) -> None:
        if self.interactive and self.lock_resources: print(f"RemoteMachineConfig Warning: If interactive is ON along with lock_resources, the job might never end.")
        if self.transfer_method == "cloud": assert self.cloud_name is not None, "Cloud name is not provided. 🤷‍♂️"
        if self.notify_upon_completion and self.to_email is None:
            from machineconfig.utils.utils import DEFAULTS_PATH
            try:
                section = tb.Read.ini(DEFAULTS_PATH)['general']
                self.to_email = section['to_email']
            except (FileNotFoundError, KeyError, IndexError) as err: raise ValueError(f"Email address is not provided. 🤷‍♂️ & default could not be read @ `{DEFAULTS_PATH}`") from err
        if self.notify_upon_completion and self.email_config_name is None:
            from machineconfig.utils.utils import DEFAULTS_PATH
            try:
                section = tb.Read.ini(DEFAULTS_PATH)['general']
                self.email_config_name = section['email_config_name']
            except (FileNotFoundError, KeyError, IndexError) as err: raise ValueError(f"Email config name is not provided. 🤷‍♂️ & default could not be read @ `{DEFAULTS_PATH}`") from err


class RemoteMachine:
    def submit_to_cloud(self, cm: CloudManager, split: int = 5, reset_cloud: bool = False) -> list['RemoteMachine']:
        """The only authority responsible for adding entries to queue df."""
        assert self.config.transfer_method == "cloud", "CloudManager only works with `transfer_method` set to `cloud`."
        assert self.config.launch_method == "cloud_manager", "CloudManager only works with `launch_method` set to `cloud_manager`."
        assert isinstance(self.ssh, SelfSSH), "CloudManager only works with `SelfSSH` objects."
        assert self.config.workload_params is None, "CloudManager only works with `workload_params` set to `None`."
        self.job_params.auto_commit()
        if reset_cloud: cm.reset_cloud()
        cm.claim_lock()  # before adding any new jobs, make sure the global jobs folder is mirrored locally.
        from copy import deepcopy
        self.config.base_dir = CloudManager.base_path.joinpath(f"jobs").collapseuser().as_posix()
        self.file_manager.base_dir = tb.P(self.config.base_dir).collapseuser()
        wl = WorkloadParams().split_to_jobs(jobs=split)
        rms: list[RemoteMachine] = []
        new_log_entries: list[LogEntry] = []
        for idx, a_workload_params in enumerate(wl):
            rm = deepcopy(self)
            rm.config.job_id = f"{rm.config.job_id}-{idx + 1}-{split}"
            rm.config.workload_params = a_workload_params
            rm.file_manager.job_root = self.file_manager.base_dir.joinpath(f"{rm.config.job_id}").collapseuser()
            rm.file_manager.job_id = rm.config.job_id
            rm.submitted = True  # must be done before generate_script which performs the pickling.
            rm.generate_scripts()
            rms.append(rm)
            new_log_entries.append(LogEntry(name=rm.config.job_id, submission_time=pd.Timestamp.now().strftime("%Y-%m-%d %H:%M:%S"), start_time=None, end_time=None, run_machine=None,
                                            source_machine=f"{getpass.getuser()}@{platform.node()}", note="", pid=None, cmd="", session_name=""))
        log = cm.read_log()  # this claims lock internally.
        new_queued_df: 'pd.DataFrame' = pd.DataFrame([item.__dict__ for item in new_log_entries])
        total_queued_df = pd.concat([log["queued"], new_queued_df], ignore_index=True, sort=False)
        log["queued"] = total_queued_df
        cm.write_log(log=log)
        cm.release_lock()  # all base_dir is synced anyway: self.resources.base_dir.joinpath(status_init).to_cloud(cloud=cm.cloud, rel2home=True)
        return rms

    def __getstate__(self) -> dict[str, Any]: return self.__dict__
    def __setstate__(self, state: dict[str, Any]): self.__dict__ = state
    def __repr__(self): return f"Compute Machine {self.ssh.get_remote_repr(add_machine=True)}"
    def __init__(self, func: Union[str, Callable[..., Any]], config: RemoteMachineConfig, func_kwargs: Optional[dict[str, Any]] = None, data: Optional[list[tb.P]] = None):
        self.config: RemoteMachineConfig = config
        self.job_params: JobParams = JobParams.from_func(func=func)
        # self.job_params.kill_at_end
        if self.config.install_repo is True: assert self.job_params.is_installabe()

        if self.config.workload_params is not None and func_kwargs is not None: assert "workload_params" not in func_kwargs, "workload_params provided twice, once in config and once in func_kwargs. 🤷‍♂️"
        self.kwargs = func_kwargs or {}
        self.data = data if data is not None else []
        # conn
        self.ssh = self.config.ssh_obj if self.config.ssh_obj is not None else tb.SSH(**self.config.ssh_params)  # type: ignore
        # scripts
        self.file_manager = FileManager(job_id=self.config.job_id, remote_machine_type=self.ssh.get_remote_machine(), base=self.config.base_dir, max_simulataneous_jobs=self.config.max_simulataneous_jobs, lock_resources=self.config.lock_resources)
        # flags
        # self.execution_command: Optional[str] = None
        self.submitted: bool = False
        self.scipts_generated: bool = False
        self.results_downloaded: bool = False
        self.results_path: Optional[tb.P] = None

    def get_session_manager(self): return Zellij() if self.ssh.get_remote_machine() != "Windows" else WindowsTerminal()
    def fire(self, run: bool = False, open_console: bool = True, launch_method: LAUNCH_METHOD = "remotely") -> tuple[int, str]:
        assert self.submitted, "Job even not submitted yet. 🤔"
        console.rule(f"Firing job `{self.config.job_id}` @ remote machine {self.ssh}")
        session_manager = self.get_session_manager()
        ssh = self.ssh
        sess_name = self.job_params.session_name
        if open_console and self.config.open_console:
            if isinstance(session_manager, Zellij):
                sess_name = session_manager.get_current_zellij_session()
                # This is a workaround that uses the same existing session and make special tab for new jobs, until zellij implements detached session capability.
                # no need to assert session started, as it is already started. Plus, The lack of suffix `sess_name (current)` creates problems.
                self.job_params.session_name = sess_name
                tb.Save.vanilla_pickle(obj=self, path=self.file_manager.remote_machine_path.expanduser(), verbose=False)
            else:
                # As for Windows Terminal, there is another problem preventing us from using the same window; there is no kill-pane or kill-tab or even kill-window, the only way is to kill process (kills window).
                # Thus, we can't terminate a job unless it has a window of its own. So we follow that apporach here.
                session_manager.open_console(sess_name=sess_name, ssh=self.ssh)
                session_manager.asssert_session_started(ssh=ssh, sess_name=sess_name)
        cmd = self.file_manager.get_fire_command(launch_method=launch_method)
        session_manager.setup_layout(ssh=ssh, sess_name=self.job_params.session_name, cmd=cmd, run=run, job_wd=self.file_manager.job_root.expanduser().absolute().as_posix(), tab_name=self.job_params.tab_name, compact=True).print()
        if isinstance(ssh, SelfSSH):
            pid_path = self.file_manager.execution_log_dir.expanduser().joinpath("pid.txt")
            while True:
                print(f"🫷 Waiting for Python process to start and declare its pid @ `{pid_path}` as dictated in python script ... ")
                time.sleep(3)
                try:
                    pid = int(pid_path.read_text())
                    import psutil
                    process_command = " ".join(psutil.Process(pid).cmdline())
                    print(f"🎉 Python process started running @ {pid=} & {process_command=}")
                    break
                except Exception: pass
        else:
            pid = 0
            process_command = "haha"
        print("\n")
        time.sleep(5)  # allow time for job to write essential log files to define itself (see execution header and repo updates lines prior to py file).
        return pid, process_command

    def run(self, run: bool = True, open_console: bool = True, show_scripts: bool = True):
        self.generate_scripts()
        if show_scripts: self.show_scripts()
        self.submit()
        self.fire(run=run, open_console=open_console)
        return self

    def submit(self) -> None:
        console.rule(title="Submitting job")
        if type(self.ssh) is SelfSSH: pass
        else:
            from crocodile.cluster.data_transfer import Submission  # import here to avoid circular import.
            if self.config.transfer_method == "transfer_sh": Submission.transfer_sh(rm=self)
            elif self.config.transfer_method == "cloud": Submission.cloud(rm=self)
            elif self.config.transfer_method == "sftp": Submission.sftp(self)
            else: raise ValueError(f"Transfer method {self.config.transfer_method} not recognized. 🤷‍")
        self.submitted = True  # before sending `self` to the remote.

    def generate_scripts(self):
        console.rule(f"Generating scripts for job `{self.file_manager.job_id}` @ Machine `{self.__repr__()}`")
        self.job_params.ssh_repr = repr(self.ssh)
        self.job_params.ssh_repr_remote = self.ssh.get_remote_repr()
        self.job_params.description = self.config.description
        self.job_params.file_manager_path = self.file_manager.file_manager_path.collapseuser().as_posix()
        self.job_params.session_name = "TS-" + tb.randstr(noun=True)  # TS: TerminalSession-CloudManager, to distinguish from other sessions created manually.
        self.job_params.tab_name = f'🏃‍♂️{self.file_manager.job_id}'  # tb.randstr(noun=True)
        execution_line = self.job_params.get_execution_line(parallelize=self.config.parallelize, workload_params=self.config.workload_params, wrap_in_try_except=self.config.wrap_in_try_except)
        py_script = tb.P(cluster.__file__).parent.joinpath("script_execution.py").read_text(encoding="utf-8").replace("params = JobParams.from_empty()", f"params = {self.job_params}").replace("# execution_line", execution_line)
        if self.config.notify_upon_completion:
            executed_obj = f"""File *{tb.P(self.job_params.repo_path_rh).joinpath(self.job_params.file_path_rh).collapseuser().as_posix()}*"""  # for email.
            assert self.config.email_config_name is not None, "Email config name is not provided. 🤷‍♂️"
            assert self.config.to_email is not None, "Email address is not provided. 🤷‍♂️"
            email_params = EmailParams(addressee=self.ssh.get_local_repr(add_machine=True),
                                       speaker=self.ssh.get_remote_repr(add_machine=True),
                                       ssh_conn_str=self.ssh.get_remote_repr(add_machine=False),
                                       executed_obj=executed_obj,
                                       file_manager_path=self.file_manager.file_manager_path.collapseuser().as_posix(),
                                       to_email=self.config.to_email, email_config_name=self.config.email_config_name)
            email_script = tb.P(cluster.__file__).parent.joinpath("script_notify_upon_completion.py").read_text(encoding="utf-8").replace("email_params = EmailParams.from_empty()", f"email_params = {email_params}").replace('manager = FileManager.from_pickle(params.file_manager_path)', '')
            py_script = py_script.replace("# NOTIFICATION-CODE-PLACEHOLDER", email_script)
        ve_path = tb.P(self.job_params.repo_path_rh).expanduser().joinpath(".ve_path")
        if ve_path.exists(): ve_name = tb.P(ve_path.read_text()).expanduser().name
        else:
            import sys
            ve_name = tb.P(sys.executable).parent.parent.name
        shell_script = f"""

# EXTRA-PLACEHOLDER-PRE

echo "~~~~~~~~~~~~~~~~SHELL START~~~~~~~~~~~~~~~"
{'~/scripts/devops -w update' if self.config.update_essential_repos else ''}
{f'cd {tb.P(self.job_params.repo_path_rh).collapseuser().as_posix()}'}
. activate_ve {ve_name}
{'git pull' if self.config.update_repo else ''}
{'pip install -e .' if self.config.install_repo else ''}
echo "~~~~~~~~~~~~~~~~SHELL  END ~~~~~~~~~~~~~~~"

echo ""
echo "Starting job {self.config.job_id} 🚀"
echo "Executing Python wrapper script: {self.file_manager.py_script_path.as_posix()}"

# EXTRA-PLACEHOLDER-POST

cd ~
{'python' if (not self.config.ipython and not self.config.pdb) else 'ipython'} {'-i' if self.config.interactive else ''} {'--pdb' if self.config.pdb else ''} {' -m pudb ' if self.config.pudb else ''} ./{self.file_manager.py_script_path.rel2home().as_posix()}

deactivate

"""  # EVERYTHING in the script above is shell-agnostic. Ensure this is the case when adding new lines.
        # shell_script_path.write_text(shell_script, encoding='utf-8', newline={"Windows": None, "Linux": "\n"}[ssh.get_remote_machine()])  # LF vs CRLF requires py3.10
        with open(file=self.file_manager.shell_script_path.expanduser().create(parents_only=True), mode='w', encoding="utf-8", newline={"Windows": None, "Linux": "\n"}[self.ssh.get_remote_machine()]) as file: file.write(shell_script)
        self.file_manager.py_script_path.expanduser().create(parents_only=True).write_text(py_script, encoding='utf-8')  # py_version = sys.version.split(".")[1]
        tb.Save.vanilla_pickle(obj=self.kwargs, path=self.file_manager.kwargs_path.expanduser(), verbose=False)
        tb.Save.vanilla_pickle(obj=self.file_manager.__getstate__(), path=self.file_manager.file_manager_path.expanduser(), verbose=False)
        tb.Save.vanilla_pickle(obj=self.config, path=self.file_manager.remote_machine_config_path.expanduser(), verbose=False)
        tb.Save.vanilla_pickle(obj=self, path=self.file_manager.remote_machine_path.expanduser(), verbose=False)
        job_status: JOB_STATUS = "queued"
        self.file_manager.execution_log_dir.expanduser().create().joinpath("status.txt").write_text(job_status)
        print("\n")

    def show_scripts(self) -> None:
        Console().print(Panel(Syntax(self.file_manager.shell_script_path.expanduser().read_text(encoding='utf-8'), lexer="ps1" if self.ssh.get_remote_machine() == "Windows" else "sh", theme="monokai", line_numbers=True), title="prepared shell script"))
        Console().print(Panel(Syntax(self.file_manager.py_script_path.expanduser().read_text(encoding='utf-8'), lexer="ps1" if self.ssh.get_remote_machine() == "Windows" else "sh", theme="monokai", line_numbers=True), title="prepared python script"))
        inspect(tb.Struct(shell_script=repr(tb.P(self.file_manager.shell_script_path).expanduser()), python_script=repr(tb.P(self.file_manager.py_script_path).expanduser()), kwargs_file=repr(tb.P(self.file_manager.kwargs_path).expanduser())), title="Prepared scripts and files.", value=False, docs=False, sort=False)

    def wait_for_results(self, sleep_minutes: int = 10) -> None:
        assert self.submitted, "Job even not submitted yet. 🤔"
        assert not self.results_downloaded, "Job already completed. 🤔"
        while True:
            tmp = self.check_job_status()
            if tmp is not None: break
            time.sleep(60 * sleep_minutes)
        self.download_results()
        if self.config.notify_upon_completion: pass

    def check_job_status(self) -> Optional[tb.P]:
        if not self.submitted:
            print("Job even not submitted yet. 🤔")
            return None
        elif self.results_downloaded:
            print("Job already completed. 🤔")
            return None

        base = self.file_manager.execution_log_dir.expanduser().create()
        try: self.ssh.copy_to_here(self.file_manager.execution_log_dir.as_posix(), z=True)
        except Exception: pass  # type: ignore  # the directory doesn't exist yet at the remote.
        end_time_file = base.joinpath("end_time.txt")

        if not end_time_file.exists():
            start_time_file = base.joinpath("start_time.txt")
            if not start_time_file.exists():
                print(f"Job {self.config.job_id} is still in the queue. 🤯")
            else:
                start_time = start_time_file.read_text()
                txt = f"Machine {self.ssh.get_remote_repr(add_machine=True)} has not yet finished job `{self.config.job_id}`. 😟"
                txt += f"\nIt started at {start_time}. 🕒, and is still running. 🏃‍♂️"
                txt += f"\nExecution time so far: {pd.Timestamp.now() - pd.to_datetime(start_time)}. 🕒"
                console.print(Panel(txt, title=f"Job `{self.config.job_id}` Status", subtitle=self.ssh.get_remote_repr(), highlight=True, border_style="bold red", style="bold"))
                print("\n")
        else:
            results_folder_file = base.joinpath("results_folder_path.txt")  # it could be one returned by function executed or one made up by the running context.
            results_folder = results_folder_file.read_text()
            print("\n" * 2)
            console.rule("Job Completed 🎉🥳🎆🥂🍾🎊🪅")
            print(f"""Machine {self.ssh.get_remote_repr(add_machine=True)} has finished job `{self.config.job_id}`. 😁
📁 results_folder_path: {results_folder} """)
            try:
                inspect(base.joinpath("execution_times.Struct.pkl").readit(), value=False, title="Execution Times", docs=False, sort=False)
            except Exception as err: print(f"Could not read execution times files. 🤷‍, here is the error:\n {err}️")
            print("\n")

            self.results_path = tb.P(results_folder)
            return self.results_path
        return None

    def download_results(self, target: Optional[str] = None, r: bool = True, zip_first: bool = False):
        assert self.results_path is not None, "Results path is unknown until job execution is finalized. 🤔\nTry checking the job status first."
        if self.results_downloaded: print(f"Results already downloaded. 🤔\nSee `{self.results_path.expanduser().absolute()}`"); return
        self.ssh.copy_to_here(source=self.results_path.collapseuser().as_posix(), target=target, r=r, z=zip_first)
        self.results_downloaded = True
        return self
    def delete_remote_results(self):
        if self.results_path is not None:
            self.ssh.run_py(cmd=f"tb.P(r'{self.results_path.as_posix()}').delete(sure=True)", verbose=False)
            return self
        else:
            print("Results path is unknown until job execution is finalized. 🤔\nTry checking the job status first.")
            return self


if __name__ == '__main__':
    # try_main()
    pass
